import logging

import click
import requests
import requests_kerberos
from click_option_group import optgroup
from click_option_group import RequiredMutuallyExclusiveOptionGroup
from ra_utils.load_settings import load_settings
from winrm.exceptions import InvalidCredentialsError

from .ad_common import AD
from .ad_logger import start_logging
from .ad_reader import ADParameterReader
from .read_ad_conf_settings import read_settings


SETTINGS = load_settings()

logger = logging.getLogger("AdTestConnectivity")


def test_basic_connectivity():
    ad = AD()
    method = ad.all_settings["primary"]["method"]
    print(f"Test basic connectivity (method={method})")
    session = ad.session

    try:
        r = session.run_cmd("ipconfig", ["/all"])
        error = None
    except requests_kerberos.exceptions.KerberosExchangeError as e:
        error = str(e)
    except InvalidCredentialsError as e:
        error = "Credentails not accepted by remote management server: {}"
        error = error.format(e)
    except requests.exceptions.ConnectionError as e:
        error = "Unable to contact winrm_host {}, message: {}"
        error = error.format(SETTINGS["integrations.ad.winrm_host"], e)

    if error is None and r.status_code == 0:
        return True
    else:
        print("Error")
        if error is None:
            print(r.std_err)
        else:
            print(error)
        return False


def test_ad_contact(index):
    try:
        ad_reader = ADParameterReader(index=index)
    except Exception as e:
        print("Unable to initiate AD-reader: {}".format(e))
        return False
    response = ad_reader.read_encoding()
    if "WindowsCodePage" in response:
        return True
    else:
        print("Error:")
        print(response)
        return False


def test_full_ad_read(index):
    """
    Test that we can read actual users from AD. This ensures suitable rights
    to access cpr-information in AD.
    """
    ad_reader = ADParameterReader(index=index)

    ad_reader.uncached_read_user(cpr="3111*")
    if ad_reader.results:
        print("Found users with bithday 31. November!")
        return False

    ad_reader.uncached_read_user(cpr="30*")
    if not ad_reader.results:
        print("No users found with bithday on the 30th of any month!")
        return False

    test_chars = {"æ": False, "ø": False, "å": False, "@": False}
    # Run through a set of users and test encoding and cpr-separator.
    for user in ad_reader.results.values():
        cprfield = ad_reader.all_settings["primary"]["cpr_field"]
        cpr = user[cprfield]
        separator = ad_reader.all_settings["primary"]["cpr_separator"]
        cpr_ok = False
        if separator == "" and len(cpr) == 10:
            cpr_ok = True
        if cpr.count(separator) == 1 and len(cpr) == 11:
            cpr_ok = True
        if not cpr.replace(separator, "").isdigit():
            cpr_ok = False
        if not cpr_ok:
            msg = "CPR-is not valid according to settings: {}, cpr-length: {}"
            print(msg.format(user["Name"], len(cpr)))
            return False

        for value in user.values():
            for char in test_chars.keys():
                if str(value).find(char) > -1:
                    test_chars[char] = True

    for test_value in test_chars.values():
        if not test_value:
            print("Finding occurences of special chars: {}".format(test_chars))
            print("(The more True the better)")
            break

    return True


def test_ad_write_settings():
    all_settings = read_settings()
    if not all_settings["primary_write"]:
        return False
    else:
        return True


def perform_read_test():
    status = "Success"
    basic_connection = test_basic_connectivity()
    if not basic_connection:
        print("Unable to connect to management server")
        exit(1)

    for index in range(len(SETTINGS["integrations.ad"])):
        print("Test AD {} contact".format(index))
        ad_connection = test_ad_contact(index)
        if not ad_connection:
            print("Unable to connect to AD {}".format(index))
            status = "Failure!!"
            continue

        print("Test ability to read from AD {}".format(index))
        full_ad_read = test_full_ad_read(index)
        if not full_ad_read:
            print("Unable to read users from AD {} correctly".format(index))
            status = "Failure!!!"
            continue

    print(status)
    # if status != "Success":
    #     exit(1)


def perform_write_test():
    ad_reader = ADParameterReader()

    print("Test that AD write settings are set up")
    write_settings = test_ad_write_settings()
    if not write_settings:
        print("Write settings not correctly set up")
        exit(1)

    # TODO: If we could make a test write, it would be nice.

    print("Write settings set up correctly")

    user_info_list = []
    i = 0
    logger.info("Check for availability of needed fields - find some test users")
    print("Check for availability of needed fields - find some test users")

    minimum_expected_fields = {
        SETTINGS["integrations.ad.write.uuid_field"]: False,
        SETTINGS["integrations.ad.write.level2orgunit_field"]: False,
        SETTINGS["integrations.ad.write.org_unit_field"]: False,
    }

    while len(user_info_list) < 10:
        i = i + 1
        if i > 31:
            msg = "Cannot find 10 users - unable to test for available fields"
            print(msg)
            logger.error(msg)
            exit(1)
        msg = "So far found {} users, current cpr: {}"
        logger.info(msg.format(len(user_info_list), str(i).zfill(2)))
        user_info = ad_reader.get_from_ad(cpr=str(i).zfill(2) + "*")
        user_info_list = user_info_list + user_info

    msg = "Found more than 10 users ({}) - now test for available fields"
    print(msg.format(len(user_info_list)))
    logger.info(msg.format(len(user_info_list)))

    for user in user_info_list:
        if isinstance(user, str):
            print("Unable to read from AD: {}".format(user))
            exit(1)

        for prop in user["PropertyNames"]:
            if prop in minimum_expected_fields:
                minimum_expected_fields[prop] = True

    if not all(minimum_expected_fields.values()):
        print("An import field is now found on the tested users")
        print(minimum_expected_fields)
    else:
        print("Test of AD fields for writing is a success")


def check_settings(settings):
    if not isinstance(settings.get("integrations.ad"), list):
        raise Exception("integrations.ad skal angives som en liste af dicts (settings)")

    if len(settings.get("integrations.ad")) < 1:
        raise Exception("integrations.ad skal indeholde mindst 1 AD (settings)")

    winrm_host_setting = "integrations.ad.winrm_host"
    winrm_host_value = settings.get(winrm_host_setting)
    if winrm_host_value is None:
        raise Exception("%r is missing" % winrm_host_setting)


@click.command(help="Test AD connectivity")
@optgroup.group("Action", cls=RequiredMutuallyExclusiveOptionGroup)
@optgroup.option("--test-read-settings", is_flag=True)
@optgroup.option("--test-write-settings", is_flag=True)
def cli(**args):
    if args.get("test_read_settings"):
        perform_read_test()
    if args.get("test_write_settings"):
        perform_read_test()
        perform_write_test()


if __name__ == "__main__":
    check_settings(SETTINGS)
    start_logging("test_connectivity.log")
    cli()



from copy import deepcopy
from datetime import date
from typing import Dict
from typing import Optional
from unittest import mock
from unittest import TestCase
from uuid import uuid4

from os2mo_helpers.mora_helpers import MoraHelper
from parameterized import parameterized

from ..config import ChangedAtSettings
from ..fix_departments import FixDepartments
from integrations.SD_Lon.tests.test_config import DEFAULT_CHANGED_AT_SETTINGS


def mock_sd_lookup(service_name, expected_params, response):
    base_responses = {
        "GetDepartment20111201": {
            "Department": [
                {
                    "DepartmentLevelIdentifier": _TestableFixDepartments.MO_CLASS_USER_KEY,
                    "DepartmentIdentifier": _TestableFixDepartments.SD_DEPARTMENT_SHORTNAME,
                    "DepartmentName": _TestableFixDepartments.SD_DEPARTMENT_NAME,
                }
            ],
        },
    }
    _response = deepcopy(base_responses[service_name])
    _response.update(response)
    sd_lookup_path = "integrations.SD_Lon.fix_departments.sd_lookup"
    sd_lookup_mock = mock.patch(sd_lookup_path, return_value=_response)
    return sd_lookup_mock


class _TestableFixDepartments(FixDepartments):
    MO_ORG_ROOT = "00000000-0000-0000-0000-000000000000"
    MO_CLASS_USER_KEY = "Enhed"
    MO_CLASS_UUID = uuid4()
    SD_INSTITUTION_UUID = uuid4()
    SD_DEPARTMENT_NAME = "some department name"
    SD_DEPARTMENT_SHORTNAME = "some department short name"
    SD_DEPARTMENT_PARENT_UUID = uuid4()

    @classmethod
    def get_instance(cls, settings_dict: Optional[Dict] = None):
        all_settings_dict = deepcopy(DEFAULT_CHANGED_AT_SETTINGS)
        if settings_dict is not None:
            all_settings_dict.update(settings_dict)
        settings = ChangedAtSettings.parse_obj(all_settings_dict)

        read_mo_org = "integrations.SD_Lon.fix_departments.MoraHelper.read_organisation"
        with mock.patch(read_mo_org, return_value=cls.MO_ORG_ROOT):
            return cls(settings)

    def get_institution(self):
        return self.SD_INSTITUTION_UUID

    def get_parent(self, unit_uuid, validity_date):
        return self.SD_DEPARTMENT_PARENT_UUID

    def _get_mora_helper(self, settings):
        mock_helper = mock.MagicMock(spec=MoraHelper)
        mock_helper.read_organisation = mock.Mock(
            return_value=_TestableFixDepartments.MO_ORG_ROOT
        )
        mock_helper.read_classes_in_facet = mock.Mock(
            return_value=[
                [{"user_key": self.MO_CLASS_USER_KEY, "uuid": self.MO_CLASS_UUID}]
            ]
        )
        return mock_helper


class TestFixDepartmentsRootSetting(TestCase):
    alternate_root = str(uuid4())

    @parameterized.expand(
        [
            # Case 1: Default root
            (
                {},
                _TestableFixDepartments.MO_ORG_ROOT,
            ),
            # Case 2: Alternate root
            (
                {
                    "sd_fix_departments_root": alternate_root,
                },
                alternate_root,
            ),
        ]
    )
    def test_root(self, settings, expected_root):
        instance = _TestableFixDepartments.get_instance(settings_dict=settings)
        self.assertEqual(instance.org_uuid, expected_root)


class TestFixDepartmentAtSingleDate(TestCase):
    def test_run(self):
        instance = _TestableFixDepartments.get_instance()
        with mock_sd_lookup("GetDepartment20111201", {}, {}):
            instance.fix_department_at_single_date("uuid", date(2020, 1, 1))
            instance.helper._mo_post.assert_called_with(
                "details/edit",
                {
                    "type": "org_unit",
                    "data": {
                        "uuid": "uuid",
                        "user_key": instance.SD_DEPARTMENT_SHORTNAME,
                        "name": instance.SD_DEPARTMENT_NAME,
                        "parent": {"uuid": instance.SD_DEPARTMENT_PARENT_UUID},
                        "org_unit_level": {"uuid": instance.MO_CLASS_UUID},
                        "org_unit_type": {"uuid": instance.MO_CLASS_UUID},
                        "validity": {"from": "1930-01-01", "to": None},
                    },
                },
            )



import json
import os
import pathlib
import time
import unittest

from integrations.os2sync import __main__, config

config.logformat = "%(message)s"


class Tests(unittest.TestCase):
    maxDiff = None

    def run_once(self, directory, cache, **overrides):
        starttime = time.time()
        pd = pathlib.Path(directory)
        log = pd / "log"
        hk = pd / "hk"
        cache = pd / cache
        config.settings["OS2SYNC_HASH_CACHE"] = str(hk)
        config.settings["MOX_LOG_FILE"] = str(log)
        # set this in settings file instead
        # config.settings["OS2SYNC_API_URL"] = "stub"
        config.settings.update(overrides)
        if not cache.exists():
            __main__.main(config.settings)
            cache.write_text(hk.read_text())
            hk.unlink()

        endtime = time.time()
        print("\ntime consumed for", directory, overrides, "is", endtime - starttime)
        return json.loads(cache.read_text())

    def test_ekvivalens(self):
        self.maxDiff = None
        testdir = "/tmp/os2sync-test"
        os.makedirs(testdir, exist_ok=True)
        old = self.run_once(testdir, "old", OS2SYNC_USE_LC_DB=False)
        new = self.run_once(testdir, "new", OS2SYNC_USE_LC_DB=True)
        self.assertEqual(old, new)



import unittest

from helpers import dummy_settings
from sqlalchemy.orm import sessionmaker

from constants import AD_it_system
from exporters.sql_export.lc_for_jobs_db import get_engine
from exporters.sql_export.sql_table_defs import Adresse
from exporters.sql_export.sql_table_defs import Base
from exporters.sql_export.sql_table_defs import Bruger
from exporters.sql_export.sql_table_defs import Engagement
from exporters.sql_export.sql_table_defs import ItForbindelse
from exporters.sql_export.sql_table_defs import ItSystem
from exporters.sql_export.sql_table_defs import Tilknytning
from integrations.os2sync.lcdb_os2mo import get_sts_user
from integrations.os2sync.lcdb_os2mo import try_get_ad_user_key


class Tests_lc_db(unittest.TestCase):
    def setUp(self):
        """
        setup db and populate with quite minimal data
        """
        self.engine = get_engine(dbpath=":memory:")
        self.session = sessionmaker(bind=self.engine, autoflush=False)()
        # Lav tables via tabledefs fra LoraCache og fyld dataen ind
        Base.metadata.create_all(self.engine)
        bruger = Bruger(
            fornavn="fornavn",
            efternavn="efternavn",
            uuid="b1",
            bvn="b1bvn",
            cpr="cpr1",
        )
        self.session.add(bruger)
        it = ItSystem(navn=AD_it_system, uuid="ItSystem1")
        self.session.add(it)
        it = ItSystem(navn="it_navn2", uuid="ItSystem2")
        self.session.add(it)
        it = ItForbindelse(
            id=1,
            uuid="if1",
            it_system_uuid="ItSystem1",
            bruger_uuid="b1",
            enhed_uuid="e1",
            brugernavn="AD-logon",
            startdato="0",
            slutdato="1",
            primær_boolean=True,
        )
        self.session.add(it)
        it = ItForbindelse(
            id=2,
            uuid="if2",
            it_system_uuid="ItSystem2",
            bruger_uuid="b1",
            enhed_uuid="e1",
            brugernavn="if_bvn2",
            startdato="0",
            slutdato="1",
            primær_boolean=True,
        )
        self.session.add(it)
        self.session.commit()

    def setup_wide(self):
        """
        setup a bunch of additional stuff
        """
        tilknytning = Tilknytning(
            uuid="t1",
            bvn="t1bvn",
            bruger_uuid="b1",
            enhed_uuid="E2",
            tilknytningstype_titel="titel",
        )
        self.session.add(tilknytning)
        engagement = Engagement(
            uuid="Eng1",
            bvn="Eng1bvn",
            engagementstype_titel="test1",
            primærtype_titel="?",
            bruger_uuid="b1",
            enhed_uuid="E3",
            stillingsbetegnelse_titel="tester1",
        )
        self.session.add(engagement)
        bruger = Bruger(
            fornavn="fornavn2",
            efternavn="efternavn2",
            uuid="b2",
            bvn="b2bvn",
            cpr="cpr2",
        )
        self.session.add(bruger)
        tilknytning = Tilknytning(
            uuid="t2",
            bvn="t2bvn",
            bruger_uuid="b2",
            enhed_uuid="E3",
            tilknytningstype_titel="titel2",
        )
        self.session.add(tilknytning)
        engagement = Engagement(
            uuid="Eng2",
            bvn="Eng2bvn",
            engagementstype_titel="test2",
            primærtype_titel="?",
            bruger_uuid="b2",
            enhed_uuid="E2",
            stillingsbetegnelse_titel="tester2",
        )
        self.session.add(engagement)
        adresse = Adresse(
            uuid="A1",
            bruger_uuid="b1",
            adressetype_scope="E-mail",
            adressetype_bvn="Email",
            adressetype_titel="Email",
            værdi="test@email.dk",
            synlighed_titel="Hemmelig",
        )
        self.session.add(adresse)
        adresse = Adresse(
            uuid="A1",
            bruger_uuid="b1",
            adressetype_scope="E-mail",
            adressetype_bvn="AD-Email",
            adressetype_titel="AD-Email",
            værdi="AD-email@email.dk",
            synlighed_titel="Offentlig",
        )
        self.session.add(adresse)
        adresse = Adresse(
            uuid="A2",
            bruger_uuid="b1",
            adressetype_scope="Telefon",
            adressetype_bvn="AD-Telefonnummer",
            adressetype_titel="AD-Telefonnummer",
            værdi="12345678",
            synlighed_titel="",
        )

        self.session.add(adresse)
        self.session.commit()

    def tearDown(self):
        Base.metadata.drop_all(self.engine)

    def test_lcdb_get_ad(self):
        expected = "AD-logon"
        self.assertEqual(expected, try_get_ad_user_key(session=self.session, uuid="b1"))

    def test_lcdb_get_sts_user_default(self):
        self.setup_wide()
        expected = {
            "Email": "test@email.dk",
            "Person": {"Cpr": "cpr1", "Name": "fornavn efternavn"},
            "PhoneNumber": "12345678",
            "UserId": "AD-logon",
            "Uuid": "b1",
            "Positions": [],
        }
        settings = dummy_settings
        settings.os2sync_xfer_cpr = True
        self.assertEqual(
            expected, get_sts_user(self.session, "b1", [], settings=settings)
        )



from unittest.mock import Mock
from uuid import uuid4

from os2mo_helpers.mora_helpers import MoraHelper
from winrm import Session

from ..ad_common import AD
from ..ad_writer import MORESTSource
from .test_utils import TestADWriterMixin

MO_ROOT_ORG_UNIT_UUID = "not-a-mo-org-unit-uuid"
MO_CHILD_ORG_UNIT_UUID = uuid4()
MO_UUID = "not-a-uuid"
AD_UUID_FIELD = "uuidField"
UNKNOWN_CPR_NO = "not-a-cpr-no"


class MockAD(AD):
    def __init__(self):
        self.session = Mock(spec=Session)
        self.all_settings = {"primary": {"search_base": ""}}


class MockADParameterReader(TestADWriterMixin):
    """Mock implementation of `ADParameterReader` which always returns the same
    AD user."""

    def read_user(self, cpr=None, **kwargs):
        return self._prepare_get_from_ad(ad_transformer=None)

    def read_it_all(self, **kwargs):
        return [self.read_user()]

    def cache_all(self, **kwargs):
        return self.read_it_all()


class MockEmptyADReader(MockADParameterReader):
    """Mock implementation of `ADParameterReader` which simulates an empty AD"""

    def read_user(self, **kwargs):
        return None

    def read_it_all(self, **kwargs):
        return []

    def cache_all(self, **kwargs):
        return self.read_it_all()


class MockUnknownCPRADParameterReader(MockADParameterReader):
    def read_user(self, cpr=None, **kwargs):
        if cpr == UNKNOWN_CPR_NO:
            return None
        return super().read_user(cpr=cpr, **kwargs)


class MockMORESTSource(MORESTSource):
    def __init__(self, from_date, to_date):
        self.from_date = from_date
        self.to_date = to_date

    def get_engagement_dates(self, uuid):
        # Return 2-tuple of (from_dates, to_dates)
        return [self.from_date], [self.to_date]


class MockLoraCache:
    # This implements enough of the real `LoraCache` to make
    # `ad_sync.AdMoSync._edit_engagement` happy.

    def __init__(self, mo_values, mo_engagements=None):
        self._mo_values = mo_values
        self._mo_engagements = mo_engagements

    @property
    def users(self):
        return {self._mo_values["uuid"]: [self._mo_values]}

    @property
    def engagements(self):
        extensions = {"udvidelse_%d" % n: "old mo value #%d" % n for n in range(1, 11)}
        if self._mo_engagements:
            return {
                eng["uuid"]: [
                    {
                        "uuid": eng["uuid"],
                        "user": self._mo_values["uuid"],
                        "primary_boolean": eng["is_primary"],
                        "from_date": eng["validity"]["from"],
                        "to_date": eng["validity"]["to"],
                        "extensions": extensions,
                    }
                ]
                for eng in self._mo_engagements
            }
        else:
            return {
                "engagement_uuid": [
                    {
                        "uuid": "engagement_uuid",
                        "user": self._mo_values["uuid"],
                        "primary_boolean": True,
                        "from_date": "1960-01-01",
                        "to_date": None,
                        "extensions": extensions,
                        # Additional keys read by `AdLifeCycle._gen_filtered_employees`
                        "job_function": None,
                        "primary_type": None,
                        "engagement_type": None,
                    }
                ]
            }


class MockLoraCacheExtended(MockLoraCache):
    """Mocks enough of `LoraCache` to test `AdLifeCycle`"""

    def populate_cache(self, **kwargs):
        pass

    def calculate_derived_unit_data(self):
        pass

    def calculate_primary_engagements(self):
        pass

    @property
    def units(self):
        # Return a single org unit (= the root org unit)
        return {
            MO_ROOT_ORG_UNIT_UUID: [
                {
                    "uuid": MO_ROOT_ORG_UNIT_UUID,
                }
            ]
        }

    @property
    def classes(self):
        return {None: {}}

    def _load_settings(self):
        return {}

    def _read_org_uuid(self):
        return "not-a-mo-org-uuid"


class MockLoraCacheEmptyEmployee(MockLoraCacheExtended):
    @property
    def users(self):
        return {self._mo_values["uuid"]: []}


class MockLoraCacheEmptyUnit(MockLoraCacheExtended):
    """Mock a LoraCache where there are no organisational units"""

    @property
    def units(self):
        return {}


class MockLoraCacheDanglingParentUnit(MockLoraCacheExtended):
    """Mock a LoraCache where organisational unit we look for has an unknown
    parent organisational unit UUID.
    """

    @property
    def units(self):
        return {
            MO_CHILD_ORG_UNIT_UUID: [
                {
                    "uuid": MO_CHILD_ORG_UNIT_UUID,
                    "parent": uuid4(),
                }
            ],
        }


class MockLoraCacheParentChildUnit(MockLoraCacheExtended):
    """Mock a LoraCache where a child unit points correctly to its parent unit
    (which is also the root unit in this case.)
    """

    @property
    def units(self):
        return {
            MO_ROOT_ORG_UNIT_UUID: [
                {
                    "uuid": MO_ROOT_ORG_UNIT_UUID,
                    "parent": None,
                }
            ],
            MO_CHILD_ORG_UNIT_UUID: [
                {
                    "uuid": MO_CHILD_ORG_UNIT_UUID,
                    "parent": MO_ROOT_ORG_UNIT_UUID,
                }
            ],
        }


class MockLoraCacheParentUnitUnset(MockLoraCacheExtended):
    """Mock a LoraCache where a child unit does not point correctly to its
    parent unit, due to its 'parent' key being None.
    """

    @property
    def units(self):
        return {
            MO_ROOT_ORG_UNIT_UUID: [
                {
                    "uuid": MO_ROOT_ORG_UNIT_UUID,
                    "parent": None,
                }
            ],
            MO_CHILD_ORG_UNIT_UUID: [
                {
                    "uuid": MO_CHILD_ORG_UNIT_UUID,
                    "parent": None,
                }
            ],
        }


class MockMoraHelper(MoraHelper):
    def __init__(self, cpr):
        self._mo_user = {"cpr_no": cpr, "uuid": MO_UUID}
        self._read_user_calls = []
        super().__init__()

    def read_organisation(self):
        return "not-a-org-uuid"

    def read_user(self, user_cpr=None, **kwargs):
        self._read_user_calls.append(user_cpr)
        if user_cpr == UNKNOWN_CPR_NO:
            return None
        return self._mo_user

    def read_all_users(self):
        return [self._mo_user]



import time
import json
import pathlib
import logging
import datetime

import click
from click_option_group import optgroup, RequiredMutuallyExclusiveOptionGroup

# SD?
# from integrations.SD_Lon import sd_payloads
from integrations.opus import payloads
from os2mo_helpers.mora_helpers import MoraHelper
from ra_utils.load_settings import load_settings


SETTINGS = load_settings()
MORA_BASE = SETTINGS['mora.base']

logger = logging.getLogger("updatePrimaryEngagements")
LOG_LEVEL = logging.DEBUG
LOG_FILE = 'calculate_primary.log'


class MOPrimaryEngagementUpdater(object):
    def __init__(self):
        self.helper = MoraHelper(hostname=MORA_BASE, use_cache=False)
        self.org_uuid = self.helper.read_organisation()

        self.mo_person = None

        # Currently primary is set first by engagement type (order given in
        # settings) and secondly by job_id. self.primary is an ordered list of
        # classes that can considered to be primary. self.primary_types is a dict
        # with all classes in the primary facet.
        self.eng_types_order = SETTINGS['integrations.opus.eng_types_primary_order']
        self.primary_types, self.primary = self._find_primary_types()

    def _find_primary_types(self):
        """
        Read the engagement types from MO and match them up against the three
        known types in the OPUS->MO import.
        :param helper: An instance of mora-helpers.
        :return: A dict matching up the engagement types with LoRa class uuids.
        """
        # These constants are global in all OPUS municipalities (because they are
        # created by the OPUS->MO importer.
        PRIMARY = 'primary'
        NON_PRIMARY = 'non-primary'
        FIXED_PRIMARY = 'explicitly-primary'

        logger.info('Read primary types')
        primary_dict = {
            'fixed_primary': None,
            'primary': None,
            'non_primary': None
        }

        primary_types = self.helper.read_classes_in_facet('primary_type')
        for primary_type in primary_types[0]:
            if primary_type['user_key'] == PRIMARY:
                primary_dict['primary'] = primary_type['uuid']
            if primary_type['user_key'] == NON_PRIMARY:
                primary_dict['non_primary'] = primary_type['uuid']
            if primary_type['user_key'] == FIXED_PRIMARY:
                primary_dict['fixed_primary'] = primary_type['uuid']

        if None in primary_dict.values():
            raise Exception('Missing primary types: {}'.format(primary_dict))
        primary_list = [primary_dict['fixed_primary'], primary_dict['primary']]

        return primary_dict, primary_list

    def set_current_person(self, cpr=None, uuid=None, mo_person=None):
        """
        Set a new person as the current user. Either a cpr-number or
        an uuid should be given, not both.
        :param cpr: cpr number of the person.
        :param uuid: MO uuid of the person.
        :param mo_person: An already existing user object from mora_helper.
        :return: True if current user is valid, otherwise False.
        """
        if uuid:
            mo_person = self.helper.read_user(user_uuid=uuid)
        elif cpr:
            mo_person = self.helper.read_user(user_cpr=cpr, org_uuid=self.org_uuid)
        elif mo_person:
            pass
        else:
            mo_person = None
        # print('Read user: {}s'.format(time.time() - t))
        if mo_person:
            self.mo_person = mo_person
            success = True
        else:
            self.mo_person = None
            success = False
        return success

    def _engagements_included_in_primary_calculation(self, engagements):
        included = []
        for eng in engagements:
            # disregard engagements from externals
            if eng["org_unit"]["uuid"] in SETTINGS.get("integrations.ad.import_ou.mo_unit_uuid",""):
                logger.warning('disregarding external engagement: {}'.format(eng))
                continue
            included.append(eng)
        return included

    def _calculate_rate_and_ids(self, mo_engagement):
        min_type_pri = 9999
        min_id = 9999999
        for eng in mo_engagement:
            logger.debug('Calculate rate, engagement: {}'.format(eng))

            try:
                employment_id = int(eng['user_key'])
            except ValueError:
                logger.warning("Skippning engangement with non-integer employment_id: {}".format(eng['user_key']))
                continue

            stat = 'Current eng_type, min_id: {}, {}. This rate, eng_pos: {}, {}'
            logger.debug(stat.format(min_type_pri, min_id,
                                     employment_id, eng['fraction']))

            if eng['engagement_type'] in self.eng_types_order:
                type_pri = self.eng_types_order.index(eng['engagement_type'])
            else:
                type_pri = 9999

            if type_pri == min_type_pri:
                if employment_id < min_id:
                    min_id = employment_id
            if type_pri < min_type_pri:
                min_id = employment_id
                min_type_pri = type_pri

        logger.debug('Min id: {}, Prioritied type: {}'.format(min_id, min_type_pri))
        return (min_id, min_type_pri)

    def check_all_for_primary(self):
        """
        Check all users for the existence of primary engagements.
        :return: TODO
        """
        count = 0
        all_users = self.helper.read_all_users()
        for user in all_users:
            if count % 250 == 0:
                print('{}/{}'.format(count, len(all_users)))
            count += 1

            self.set_current_person(uuid=user['uuid'])
            date_list = self.helper.find_cut_dates(user['uuid'])
            for i in range(0, len(date_list) - 1):
                date = date_list[i]
                mo_engagement = self.helper.read_user_engagement(
                    user=self.mo_person['uuid'], at=date, only_primary=True
                )

                primary_count = 0
                for eng in mo_engagement:
                    if eng['primary']['uuid'] in self.primary:
                        primary_count += 1
            if primary_count == 0:
                print('No primary for {} at {}'.format(user['uuid'], date))
            elif primary_count > 1:
                print('Too many primaries for {} at {}'.format(user['uuid'], date))
            else:
                # print('Correct')
                pass

    def recalculate_primary(self, no_past=True):
        """
        Re-calculate primary engagement for the entire history of the current user.
        """
        logger.info('Calculate primary engagement: {}'.format(self.mo_person))
        date_list = self.helper.find_cut_dates(self.mo_person['uuid'],
                                               no_past=no_past)

        number_of_edits = 0

        for i in range(0, len(date_list) - 1):
            date = date_list[i]
            logger.info('Recalculate primary, date: {}'.format(date))

            mo_engagement = self.helper.read_user_engagement(
                user=self.mo_person['uuid'], at=date, only_primary=True, use_cache=False
            )

            mo_engagement = self._engagements_included_in_primary_calculation(mo_engagement)
            if len(mo_engagement) == 0:
                continue

            logger.debug('MO engagement: {}'.format(mo_engagement))

            (min_id, min_type_pri) = self._calculate_rate_and_ids(mo_engagement)
            if (min_id is None) or (min_type_pri is None):
                # continue
                raise Exception('Cannot calculate primary')

            fixed = None
            for eng in mo_engagement:
                if eng['primary']:
                    if eng['primary']['uuid'] == self.primary_types['fixed_primary']:
                        logger.info('Engagment {} is fixed primary'.format(eng['uuid']))
                        fixed = eng['uuid']

            exactly_one_primary = False
            for eng in mo_engagement:
                to = datetime.datetime.strftime(
                    date_list[i + 1] - datetime.timedelta(days=1), "%Y-%m-%d"
                )
                if date_list[i + 1] == datetime.datetime(9999, 12, 30, 0, 0):
                    to = None
                validity = {
                    'from': datetime.datetime.strftime(date, "%Y-%m-%d"),
                    'to': to
                }

                try:
                    employment_id = int(eng['user_key'])
                except ValueError:
                    logger.warning("Skippning engangement with non-integer employment_id: {}".format(eng['user_key']))
                    continue
                if eng['engagement_type'] in self.eng_types_order:
                    type_pri = self.eng_types_order.index(eng['engagement_type'])
                else:
                    type_pri = 9999

                msg = 'Current type pri and id: {}, {}'
                logger.debug(msg.format(type_pri, employment_id))

                if type_pri == min_type_pri and employment_id == min_id:
                    assert(exactly_one_primary is False)
                    logger.debug('Primary is: {}'.format(employment_id))
                    exactly_one_primary = True
                    current_type = self.primary_types['primary']
                else:
                    logger.debug('{} is not primary'.format(employment_id))
                    current_type = self.primary_types['non_primary']

                if fixed is not None and eng['uuid'] != fixed:
                    # A fixed primary exits, but this is not it.
                    logger.debug('Manual override, this is not primary!')
                    current_type = self.primary_types['non_primary']
                if eng['uuid'] == fixed:
                    # This is a fixed primary.
                    current_type = self.primary_types['fixed_primary']

                data = {
                    'primary': {'uuid': current_type},
                    'validity': validity
                }

                payload = payloads.edit_engagement(data, eng['uuid'])
                if not payload['data']['primary'] == eng['primary']:
                    logger.debug('Edit payload: {}'.format(payload))
                    response = self.helper._mo_post('details/edit', payload)
                    assert response.status_code == 200
                    number_of_edits += 1
                else:
                    logger.debug('No edit, primary type not changed.')
        return_dict = {self.mo_person['uuid']: number_of_edits}
        return return_dict

    def recalculate_all(self, no_past=False):
        """
        Recalculate all primary engagements
        :return: TODO
        """
        all_users = self.helper.read_all_users()
        edit_status = {}
        for user in all_users:
            t = time.time()
            self.set_current_person(uuid=user['uuid'])
            status = self.recalculate_primary(no_past=no_past)
            edit_status.update(status)
            logger.debug('Time for primary calculation: {}'.format(time.time() - t))
        print('Total edits: {}'.format(sum(edit_status.values())))


@click.command(help="Calculate Primary")
@optgroup.group("Action", cls=RequiredMutuallyExclusiveOptionGroup)
@optgroup.option(
    "--check-all-for-primary",
    is_flag=True,
    help='Check all users for a primary engagement',
)
@optgroup.option("--recalculate-all", is_flag=True, help='Recalculate all users')
@optgroup.option(
    "--recalculate-user",
    help='Recalculate primaries for a user (specify MO UUID)',
)
def cli(**args):
    updater = MOPrimaryEngagementUpdater()

    if args['recalculate_user']:
        print('Recalculate user')
        t = time.time()
        updater.set_current_person(uuid=args['recalculate_user'])
        updater.recalculate_primary()
        print('Time for primary calculation: {}'.format(time.time() - t))

    if args['check_all_for_primary']:
        print('Check all for primary')
        updater.check_all_for_primary()

    if args['recalculate_all']:
        print('Check all for primary')
        updater.recalculate_all(no_past=True)


if __name__ == '__main__':
    detail_logging = ('mora-helper', 'updatePrimaryEngagements', 'sdCommon')
    for name in logging.root.manager.loggerDict:
        if name in detail_logging:
            logging.getLogger(name).setLevel(LOG_LEVEL)
        else:
            logging.getLogger(name).setLevel(logging.ERROR)

    logging.basicConfig(
        format='%(levelname)s %(asctime)s %(name)s %(message)s',
        level=LOG_LEVEL,
        filename=LOG_FILE
    )

    cli()


import asyncio
import datetime
from pathlib import Path

import click
import config
import initial
from dateutil import parser
from los_leder import ManagerImporter
from los_org import OrgUnitImporter
from los_pers import PersonImporter
from los_stam import StamImporter


def get_or_create_import_state(settings: config.Settings) -> datetime.datetime:
    """
    Ensure that the state file exists, and return the date of the last import
    If no import has been run, datetime.min is returned
    """
    state_file_path = Path(settings.import_state_file)
    if not state_file_path.is_file():
        print("Import file not present. Creating.")
        with open(state_file_path, "w") as f:
            earliest = datetime.datetime.min
            f.write(earliest.isoformat())
            return earliest
    else:
        with open(state_file_path, "r") as f:
            last_import = parser.parse(f.read())
            print(f"Last import performed at {last_import}")
            return last_import


def set_import_state(settings: config.Settings, import_date: datetime.datetime):
    """Set contents of import state file to specified date"""
    state_file_path = Path(settings.import_state_file)
    with open(state_file_path, "w") as f:
        import_date_string = import_date.isoformat()
        print(f"Writing timestamp '{import_date_string}' to state")
        f.write(import_date_string)


def run_los_import(settings, last_import):
    loop = asyncio.get_event_loop()

    initial_import = asyncio.ensure_future(initial.perform_initial_setup())
    loop.run_until_complete(initial_import)

    # Import STAM
    stam_import = asyncio.ensure_future(StamImporter().run(last_import))
    loop.run_until_complete(stam_import)

    # Import Org
    org_import = asyncio.ensure_future(OrgUnitImporter().run(last_import))
    loop.run_until_complete(org_import)

    # # Import Person
    person_import = asyncio.ensure_future(PersonImporter().run(last_import))
    loop.run_until_complete(person_import)

    # Import manager
    manager_import = asyncio.ensure_future(ManagerImporter().run(last_import))
    loop.run_until_complete(manager_import)

    loop.close()

    set_import_state(settings, datetime.datetime.now())


@click.command()
@click.option("--import-from-date")
@click.option(
    "--ftp-url",
    help="URL of FTP where CSV files will be retrieved from",
)
@click.option(
    "--ftp-user",
    help="Username to use when logging into FTP server",
)
@click.option(
    "--ftp-pass",
    help="Password to use when logging into FTP server",
)
@click.option(
    "--ftp-folder",
    help="FTP folder where CSV files are retrieved from",
)
@click.option(
    "--import-state-file",
    help="Name of import state file",
)
@click.option(
    "--import-csv-folder",
    help="Path to folder containing CSV files to import. Disables FTP reading",
)
@click.option(
    "--azid-it-system-uuid",
    type=click.UUID,
    help="UUID of MO IT system used for the `AZID` column of `Pers_*.csv` files",
)
def main(**kwargs):
    import_from_date = kwargs.pop("import_from_date", None)
    command_line_options = {key: value for key, value in kwargs.items() if value}
    settings = config.Settings.from_kwargs(**command_line_options)

    if import_from_date:
        last_import = datetime.date.fromisoformat(import_from_date)
    else:
        last_import = get_or_create_import_state(settings)

    return run_los_import(settings, last_import)


if __name__ == "__main__":
    main()  # type: ignore



import unittest
from collections import OrderedDict
from copy import deepcopy
from pathlib import Path
from unittest import TestCase

from hypothesis import given
from hypothesis.strategies import text
from parameterized import parameterized

from integrations.opus import opus_helpers


testfile1 = Path.cwd() / "integrations/opus/tests/ZLPE20200101_delta.xml"
testfile2 = Path.cwd() / "integrations/opus/tests/ZLPE20200102_delta.xml"


class test_opus_helpers(TestCase):
    @given(text())
    def test_generate_uuid(self, value):
        uuid1 = opus_helpers.generate_uuid(value)
        uuid2 = opus_helpers.generate_uuid(value)
        self.assertEqual(uuid1, uuid2)

    @parameterized.expand(
        [
            (testfile1,),
            (testfile2,),
        ]
    )
    def test_file_diff_same(self, file):
        file_diffs = opus_helpers.file_diff(file, file, disable_tqdm=True)
        self.assertEqual(file_diffs["units"], [])
        self.assertEqual(file_diffs["employees"], [])

    def test_file_diff(self):
        file_diffs = opus_helpers.file_diff(testfile1, testfile2, disable_tqdm=True)
        self.assertNotEqual(file_diffs["units"], [])
        self.assertNotEqual(file_diffs["employees"], [])

    @parameterized.expand(
        [
            (testfile1,),
            (testfile2,),
        ]
    )
    def test_parser(self, file):
        self.units, self.employees = opus_helpers.parser(file)
        self.assertIsInstance(self.units, list)
        self.assertIsInstance(self.employees, list)
        self.assertIsInstance(self.units[0], OrderedDict)
        self.assertIsInstance(self.employees[0], OrderedDict)

    @parameterized.expand(
        [
            (testfile1, [], set()),
            (testfile2, [], set()),
            (testfile1, ["1"], {"1", "2", "3"}),
            (testfile2, ["1"], {"1", "2", "3", "4", "5"}),
        ]
    )
    def test_all_filtered_ids_fiter_top_unit(self, file, filter_ids, expected):
        all_filtered = opus_helpers.find_all_filtered_ids(file, filter_ids)
        self.assertEqual(len(all_filtered), len(expected))
        self.assertEqual(all_filtered, expected)

    @parameterized.expand(
        [
            (testfile1, 1),
            (testfile2, 2),
        ]
    )
    def test_split_employees_active(self, inputfile, expected):
        file_diffs = opus_helpers.file_diff(None, inputfile, disable_tqdm=True)
        employees, leaves = opus_helpers.split_employees_leaves(file_diffs["employees"])
        employees = list(employees)
        leaves = list(leaves)
        self.assertEqual(len(leaves), expected)

    @parameterized.expand(
        [
            ([], [3, 0, 4, 1], None),
            (["1"], [0, 3, 0, 1], None),
            # Read specific opus_id
            ([], [1, 0, 0, 0], 1),
            (["1"], [0, 1, 0, 0], 1),
        ]
    )
    def test_full_(self, filter_ids, expected, opus_id):
        data = opus_helpers.read_and_transform_data(
            None,
            testfile1,
            filter_ids,
            disable_tqdm=True,
            opus_id=opus_id,
        )
        # data is a tuple of units, filtered units, employess, terminated employees
        # test that the length of each is as expected
        for i, x in enumerate(data):
            this = list(x)
            assert (
                len(this) == expected[i]
            ), f"Expected {expected[i]} objects, got {len(this) }\n{this}"

    def test_find_changed_parent(self):
        org1, _ = opus_helpers.parser(testfile1)
        org2 = deepcopy(org1)
        org2[2]["parentOrgUnit"] = "CHANGED!"
        diffs = opus_helpers.find_changes(before=org1, after=org2, disable_tqdm=True)
        self.assertEqual(diffs, [org2[2]])

    def test_find_cancelled(self):
        file_diffs = opus_helpers.file_diff(testfile1, testfile2, disable_tqdm=True)
        assert (
            len(file_diffs["cancelled_employees"]) == 1
        ), "Expected to find 1 cancelled employee"


if __name__ == "__main__":
    unittest.main()



from unittest.mock import patch

from integrations.os2sync.config import get_os2sync_settings

# Create dummy settings ignoring any settings.json file.
with patch("integrations.os2sync.config.load_settings", return_value={}):
    dummy_settings = get_os2sync_settings(
        municipality="1234",
        os2sync_top_unit_uuid="269a0339-0c8b-472d-9514-aef952a2b4df",
    )

NICKNAME_TEMPLATE = "{% if nickname -%}{{ nickname }}{%- else %}{{ name }}{%- endif %}"


class MoEmployeeMixin:
    def mock_employee(self, cpr="0101012222", nickname=False):
        # Mock the result of `os2mo_get("{BASE}/e/" + uuid + "/").json()`
        # Only contains the keys relevant for testing
        return {
            # Name
            "name": "Test Testesen",
            "givenname": "Test",
            "surname": "Testesen",
            # Nickname
            "nickname": "Kalde Navn" if nickname else "",
            "nickname_givenname": "Kalde" if nickname else "",
            "nickname_surname": "Navn" if nickname else "",
            # Other fields
            "cpr_no": cpr,
            "user_key": "testtestesen",
            "uuid": "mock-uuid",
        }

    def mock_employee_response(self, **kwargs):
        mo_employee = self.mock_employee(**kwargs)

        class MockResponse:
            def json(self):
                return mo_employee

        return MockResponse()



import asyncio
import time
from datetime import datetime
from operator import itemgetter
from pathlib import Path
from typing import Optional

import click
import requests
from click_option_group import MutuallyExclusiveOptionGroup
from click_option_group import optgroup
from more_itertools import first
from more_itertools import flatten
from more_itertools import pairwise
from more_itertools import partition
from more_itertools import prepend
from ra_utils.load_settings import load_settings
from tqdm import tqdm

import constants
from integrations.ad_integration import ad_reader
from integrations.opus import opus_helpers
from integrations.opus.opus_diff_import import import_one
from tools.data_fixers.class_tools import find_duplicates_classes
from tools.default_mo_setup import create_new_root_and_it
from tools.default_mo_setup import ensure_default_classes
from tools.subtreedeleter import subtreedeleter_helper


def truncate_db(MOX_BASE: str = "http://localhost:8080") -> None:
    r = requests.get(MOX_BASE + "/db/truncate")
    r.raise_for_status()


def find_opus_name() -> str:
    """Generates uuid for opus root.

    Reads the first available opus file and generates the uuid for the first unit in the file.
    Assumes this is the root organisation of opus.
    """
    dumps = opus_helpers.read_available_dumps()

    first_date = min(sorted(dumps.keys()))
    units, _ = opus_helpers.parser(dumps[first_date])
    main_unit = first(units)
    calculated_uuid = opus_helpers.generate_uuid(main_unit["@id"])
    return str(calculated_uuid)


def prepare_re_import(
    settings: Optional[list] = None,
    opus_uuid: Optional[str] = None,
    truncate: Optional[bool] = None,
    connections: int = 4,
) -> None:
    """Create a MO setup with necessary classes.

    Clear MO database, or only the opus-unit with the given uuid.
    Ensure necessary classes exists.
    """
    settings = settings or load_settings()
    mox_base = settings.get("mox.base")
    if truncate:
        truncate_db(mox_base)
        # Create root org and it systems
        create_new_root_and_it()
    elif opus_uuid:
        session = requests.session()
        dub = find_duplicates_classes(session=session, mox_base=mox_base)
        if dub:
            raise Exception(
                "There are duplicate classes, remove them with tools/data_fixers/remove_duplicate_classes.py --delete"
            )
        subtreedeleter_helper(
            opus_uuid,
            delete_functions=True,
            keep_functions=["KLE", "Relateret Enhed"],
            connections=connections,
        )
    ensure_default_classes()


def import_opus(
    ad_reader=None,
    import_all: bool = False,
    import_last=False,
    opus_id=None,
    rundb_write: bool = True,
) -> None:
    """Import one or all files from opus even if no previous files have been imported"""
    settings = load_settings()
    filter_ids = settings.get("integrations.opus.units.filter_ids", [])
    skip_employees = settings.get("integrations.opus.skip_employees", False)
    dumps = opus_helpers.read_available_dumps()

    all_dates = dumps.keys()
    # Default is read first file only
    export_dates = [min(all_dates)]
    if import_last:
        export_dates = [max(all_dates)]
    elif import_all:
        export_dates = sorted(all_dates)

    export_dates = prepend(None, export_dates)
    date_pairs = pairwise(export_dates)
    for date1, date2 in date_pairs:
        import_one(
            ad_reader,
            date2,
            date1,
            dumps,
            filter_ids,
            opus_id=opus_id,
            rundb_write=rundb_write,
        )


@click.command()
@optgroup.group(
    "Import Amount",
    cls=MutuallyExclusiveOptionGroup,
    help="Default: Imports first file",
)
@optgroup.option(
    "--import-all",
    is_flag=True,
    help="Import all opus files.",
)
@optgroup.option(
    "--import-last",
    is_flag=True,
    help="Import only latest file.",
)
@optgroup.group(
    "Clear data", cls=MutuallyExclusiveOptionGroup, help="Default: No data deleted"
)
@optgroup.option("--truncate", is_flag=True, help="Truncate all MO tables")
@optgroup.option(
    "--delete-opus",
    is_flag=True,
    help="Delete Opus subtree. Deletes all units, engagements etc, but not KLE and related units",
)
@click.option(
    "--connections",
    default=4,
    help="The amount of concurrent requests made to OS2mo. Only applies to '--delete-opus'",
)
@click.option("--new-rundb", is_flag=True, help="Initialize new run-db")
@click.option("--use-ad", is_flag=True, help="Read from AD")
def clear_and_reload(
    import_all: bool,
    import_last: bool,
    delete_opus: bool,
    truncate: bool,
    new_rundb: bool,
    use_ad: bool,
    connections: int,
) -> None:
    """Tool for reimporting opus files.

    This tool will load the first opus-file into MO if no inputs are given.
    If the opus organisation allready exists, use either --delete-opus or --truncate.
    --delete-opus will use an uuid genereted from the first unit of the first opus-file and delete all units under this unit, and all relations to thoose units except KLE and related units.
    --truncate will truncate the database entirely.
    Add the --use-ad flag to connect to AD when reading users.
    """
    settings = load_settings()
    if truncate:
        click.confirm(
            "This will purge ALL DATA FROM MO. Do you want to continue?", abort=True
        )
    if new_rundb:
        opus_helpers.initialize_db(settings["integrations.opus.import.run_db"])

    opus_uuid = find_opus_name() if delete_opus else None
    prepare_re_import(
        settings=settings,
        opus_uuid=opus_uuid,
        truncate=truncate,
        connections=connections,
    )
    AD = None
    if use_ad:
        AD = ad_reader.ADParameterReader()
        AD.cache_all(print_progress=True)
    import_opus(ad_reader=AD, import_all=import_all, import_last=import_last)


if __name__ == "__main__":
    clear_and_reload()



import unittest

from integrations.os2sync.os2mo import addresses_to_orgunit


class _AddressMixin:
    def mock_address_list(self, scope, user_key, value):
        # Mock the result of
        # `os2mo_get("{BASE}/ou/" + uuid + "/details/address").json()`
        # Only contains the keys relevant for `addresses_to_orgunit`
        return [
            {
                "address_type": {
                    "scope": scope,
                    "user_key": user_key,
                },
                "name": value,
            }
        ]


class TestContactOpenHours(unittest.TestCase, _AddressMixin):
    def test_contact_open_hours(self):
        result = {}
        mo_data = self.mock_address_list(
            "TEXT", "ContactOpenHours", "Man-fre: 11-13.30"
        )
        addresses_to_orgunit(result, mo_data)  # Mutates `result`
        self.assertDictEqual(result, {"ContactOpenHours": "Man-fre: 11-13.30"})


class TestDtrId(unittest.TestCase, _AddressMixin):
    def test_dtr_id(self):
        result = {}
        mo_data = self.mock_address_list("TEXT", "DtrId", "G123456")
        addresses_to_orgunit(result, mo_data)  # Mutates `result`
        self.assertDictEqual(result, {"DtrId": "G123456"})



# TODO:
# All templates might need to also inlclude bp['server'],
# no relevant test-case is currently available.
# Set password for account
set_password_template = """
Get-ADUser -Filter 'SamAccountName -eq \"{username}\"'
           -Credential $usercredential |

Set-ADAccountPassword
   -Reset
   -NewPassword (ConvertTo-SecureString
                 -AsPlainText "{password}" -Force)
    -Credential $usercredential
"""


# Enable AD account
enable_user_template = """
Get-ADUser -Filter 'SamAccountName -eq \"{username}\"'
            -Credential $usercredential |
Enable-ADAccount -Credential $usercredential
"""

# Disable AD account
disable_user_template = """
Get-ADUser -Filter 'SamAccountName -eq \"{username}\"'
            -Credential $usercredential |
Disable-ADAccount -Credential $usercredential
"""


# Delete user
delete_user_template = """
Get-ADUser -Filter 'SamAccountName -eq \"{username}\"' -Credential $usercredential |
Remove-ADUser -Credential $usercredential -Confirm:$false
"""


# Add manager to user
add_manager_template = """
Get-ADUser -Filter 'SamAccountName -eq \"{user_sam}\"' -Credential $usercredential |
Set-ADUser -Manager {manager_sam} -Credential $usercredential
"""


rename_user_template = """
Get-ADUser -Filter 'SamAccountName -eq \"{user_sam}\"' -Credential $usercredential |
Rename-ADobject -Credential $usercredential -NewName "{new_name}"
"""



from uuid import UUID

import click

from integrations.os2sync.os2sync import delete_user


@click.command()
@click.option(
    "--uuid",
    type=click.UUID,
    required=True,
    help="UUID of the MO user to delete in FK org.",
)
def delete_fk_org_user(uuid: UUID) -> None:
    """Delete a MO user in FK org."""
    delete_user(str(uuid))


if __name__ == "__main__":
    delete_fk_org_user()



from datetime import datetime

import los_files
import los_stam
from hypothesis import given
from pydantic import Field

from .helpers import HelperMixin
from .strategies import csv_buf_from_model


class MockStamCSV(los_stam.StamCSV):
    mock_field: str = Field(default="testing testing")

    @staticmethod
    def get_filename() -> str:
        return "mock-filename.csv"


class TestStamImporterLoadCSVIfNewer(HelperMixin):
    """Test `los_stam.StamImporter._load_csv_if_newer`."""

    _datetime_last_imported = datetime(2020, 1, 1)
    _datetime_last_modified = datetime(2020, 1, 2)

    def test_returns_csv_if_file_is_newer(self):
        # Pretend that `get_modified_datetime_for_file` finds a file, and that
        # `read_csv` returns its parsed contents.
        with self._mock_get_modified_datetime(
            return_value=self._datetime_last_modified
        ):
            with self._mock_read_csv(MockStamCSV()):
                result = self._run_load_csv_if_newer()
                assert len(result) == 1
                assert isinstance(result[0], MockStamCSV)
                assert result[0].mock_field == MockStamCSV().mock_field

    def test_returns_nothing_if_file_is_older(self):
        # Pretend that `get_modified_datetime_for_file` finds a file, but its
        # modified date is equal to the date of the last import run. In that
        # case, return nothing.
        with self._mock_get_modified_datetime(
            return_value=self._datetime_last_imported
        ):
            result = self._run_load_csv_if_newer()
            assert result is None

    def test_handles_missing_file(self):
        # Pretend that no file is found by `get_modified_datetime_for_file`.
        with self._mock_get_modified_datetime(side_effect=ValueError()):
            result = self._run_load_csv_if_newer()
            assert result is None

    def _run_load_csv_if_newer(self):
        return los_stam.StamImporter._load_csv_if_newer(
            MockStamCSV, self._datetime_last_imported
        )


class TestParseEngagementstypeCSV:
    @given(csv_buf_from_model(model=los_stam.Engagementstype))
    def test_parse_raises_nothing(self, csv_buf):
        los_files.parse_csv(csv_buf.readlines(), los_stam.Engagementstype)


class TestParseEnhedstypeCSV:
    @given(csv_buf_from_model(model=los_stam.Enhedstype))
    def test_parse_raises_nothing(self, csv_buf):
        los_files.parse_csv(csv_buf.readlines(), los_stam.Enhedstype)


class TestParseITSystemCSV:
    @given(csv_buf_from_model(model=los_stam.ITSystem))
    def test_parse_raises_nothing(self, csv_buf):
        los_files.parse_csv(csv_buf.readlines(), los_stam.ITSystem)


class TestParseStillingsbetegnelseCSV:
    @given(csv_buf_from_model(model=los_stam.Stillingsbetegnelse))
    def test_parse_raises_nothing(self, csv_buf):
        los_files.parse_csv(csv_buf.readlines(), los_stam.Stillingsbetegnelse)


class TestParseLederansvarCSV:
    @given(csv_buf_from_model(model=los_stam.Lederansvar))
    def test_parse_raises_nothing(self, csv_buf):
        los_files.parse_csv(csv_buf.readlines(), los_stam.Lederansvar)


class TestParseLederniveauCSV:
    @given(csv_buf_from_model(model=los_stam.Lederniveau))
    def test_parse_raises_nothing(self, csv_buf):
        los_files.parse_csv(csv_buf.readlines(), los_stam.Lederniveau)


class TestParseLedertypeCSV:
    @given(csv_buf_from_model(model=los_stam.Ledertype))
    def test_parse_raises_nothing(self, csv_buf):
        los_files.parse_csv(csv_buf.readlines(), los_stam.Ledertype)


class TestParseTilknytningsrolleCSV:
    @given(csv_buf_from_model(model=los_stam.Tilknytningsrolle))
    def test_parse_raises_nothing(self, csv_buf):
        los_files.parse_csv(csv_buf.readlines(), los_stam.Tilknytningsrolle)



import unittest
from unittest.mock import patch

from alchemy_mock.mocking import UnifiedAlchemyMagicMock
from helpers import dummy_settings
from parameterized import parameterized

from exporters.sql_export.sql_table_defs import Bruger
from integrations.os2sync import lcdb_os2mo
from integrations.os2sync.tests.helpers import NICKNAME_TEMPLATE

# Mock contents of `Bruger` model
_lcdb_mock_users = [
    (
        # When this query occurs:
        [unittest.mock.call.filter(Bruger.uuid == "name only")],
        # Return this object:
        [
            Bruger(
                fornavn="Test",
                efternavn="Testesen",
                kaldenavn_fornavn="",
                kaldenavn_efternavn="",
                bvn="mo-user-key",
            ),
        ],
    ),
    (
        # When this query occurs:
        [unittest.mock.call.filter(Bruger.uuid == "name and nickname")],
        # Return this object:
        [
            Bruger(
                fornavn="Test",
                efternavn="Testesen",
                kaldenavn_fornavn="Teste",
                kaldenavn_efternavn="Testersen",
                bvn="mo-user-key",
            ),
        ],
    ),
]


class TestGetStsUser(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self._session = UnifiedAlchemyMagicMock(data=_lcdb_mock_users)

    @parameterized.expand(
        [
            # Test without template
            (
                None,  # template
                "name only",  # uuid of mock `Bruger`
                "Test Testesen",  # expected value of `Name`
            ),
            # Test with template: user has no nickname
            (
                NICKNAME_TEMPLATE,  # template
                "name only",  # uuid of mock `Bruger`
                "Test Testesen",  # expected value of `Name`
            ),
            # Test with template: user has no nickname
            (
                NICKNAME_TEMPLATE,  # template
                "name and nickname",  # uuid of mock `Bruger`
                "Teste Testersen",  # expected value of `Name`
            ),
        ]
    )
    def test_person_template_nickname(self, template, uuid, expected_name):
        settings = dummy_settings
        if template:
            # Run with template
            settings.os2sync_templates["person.name"] = template
            sts_user = lcdb_os2mo.get_sts_user(
                self._session, uuid, [], settings=settings
            )
        else:
            # Run without template
            sts_user = lcdb_os2mo.get_sts_user(
                self._session, uuid, [], settings=settings
            )

        self.assertDictEqual(
            sts_user,
            {
                "Uuid": uuid,
                "UserId": uuid,
                "Positions": [],
                "Person": {
                    "Name": expected_name,
                    "Cpr": None,
                },
            },
        )

    @parameterized.expand(
        [
            # Test without an AD BVN and without template
            (
                None,  # template config
                None,  # return value of `try_get_ad_user_key`
                "name only",  # expected value of `UserId` (MO UUID)
            ),
            # Test without an AD BVN, and template which uses `user_key`
            (
                {"person.user_id": "{{ user_key }}"},  # template config
                None,  # return value of `try_get_ad_user_key`
                "mo-user-key",  # expected value of `UserId` (MO BVN)
            ),
            # Test without an AD BVN, and template which uses `uuid`
            (
                {"person.user_id": "{{ uuid }}"},  # template config
                None,  # return value of `try_get_ad_user_key`
                "name only",  # expected value of `UserId` (MO UUID)
            ),
            # Test with an AD BVN, but without template
            (
                None,  # template config
                "mock-ad-bvn",  # return value of `try_get_ad_user_key`
                "mock-ad-bvn",  # expected value of `UserId` (AD BVN)
            ),
            # Test with an AD BVN, and template which uses `user_key`
            (
                {"person.user_id": "{{ user_key }}"},  # template config
                "mock-ad-bvn",  # return value of `try_get_ad_user_key`
                "mock-ad-bvn",  # expected value of `UserId` (AD BVN)
            ),
            # Test with an AD BVN, and template which uses `uuid`
            (
                {"person.user_id": "{{ uuid }}"},  # template config
                "mock-ad-bvn",  # return value of `try_get_ad_user_key`
                "mock-ad-bvn",  # expected value of `UserId` (AD BVN)
            ),
        ]
    )
    def test_user_template_user_id(
        self,
        os2sync_templates,
        given_ad_user_key,
        expected_user_id,
    ):
        mo_user_uuid = "name only"
        settings = dummy_settings
        settings.os2sync_templates = os2sync_templates or {}
        with self._patch("try_get_ad_user_key", given_ad_user_key):
            sts_user = lcdb_os2mo.get_sts_user(
                self._session, mo_user_uuid, [], settings=dummy_settings
            )

        self.assertDictEqual(
            sts_user,
            {
                "Uuid": mo_user_uuid,
                "UserId": expected_user_id,
                "Positions": [],
                "Person": {
                    "Name": "Test Testesen",
                    "Cpr": None,
                },
            },
        )

    def _patch(self, name, return_value):
        return patch.object(lcdb_os2mo, name, return_value=return_value)



# -- coding: utf-8 --
#
# Copyright (c) 2018, Magenta ApS
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import hashlib
import json
import logging

import requests

from integrations.os2sync import config

settings = config.get_os2sync_settings()
logger = logging.getLogger(config.loggername)
hash_cache = {}
session = requests.Session()


if settings.os2sync_api_url == "stub":
    from integrations.os2sync import stub
    session = stub.Session()


session.verify = settings.os2sync_ca_verify_os2sync
session.headers = {
    "User-Agent": "os2mo-data-import-and-export",
    "CVR": settings.municipality
}


def already_xferred(url, params, method):
    if settings.os2sync_api_url == "stub":
        params_hash = params
    else:
        params_hash = hashlib.sha224(
            (json.dumps(params, sort_keys=True) + method).encode("utf-8")
        ).hexdigest()
    if hash_cache.get(url) == params_hash:
        return True
    else:
        hash_cache[url] = params_hash
    return False


def os2sync_url(url):
    """format url like {BASE}/user
    """
    url = url.format(BASE=settings.os2sync_api_url)
    return url


def os2sync_get(url, **params):
    url = os2sync_url(url)
    try:
        r = session.get(url, params=params)
        r.raise_for_status()
        return r
    except Exception:
        logger.error(url + " " + r.text)
        logger.exception("")
        raise


def os2sync_delete(url, **params):
    url = os2sync_url(url)
    try:
        r = session.delete(url, **params)
        r.raise_for_status()
        return r
    except requests.HTTPError as e:
        if e.response.status_code == 404:
            logger.warning("delete %r %r :404", url, params)
            return r
    except Exception:
        logger.error(url + " " + r.text)
        logger.exception("")
        raise


def os2sync_post(url, **params):
    url = os2sync_url(url)
    try:
        r = session.post(url, **params)
        r.raise_for_status()
        return r
    except Exception:
        logger.error(url + " " + r.text)
        logger.exception("")
        raise


def user_uuids():
    return os2sync_get("{BASE}/user").json()


def delete_user(uuid):
    if not already_xferred("/user/" + uuid, {}, "delete"):
        logger.info("delete user %s", uuid)
        os2sync_delete("{BASE}/user/" + uuid)
    else:
        logger.info("delete user %s - cached", uuid)


def upsert_user(user):
    if not already_xferred("/user/" + user["Uuid"], user, "upsert"):
        logger.info("upsert user %s", user["Uuid"])
        os2sync_post("{BASE}/user", json=user)
    else:
        logger.info("upsert user %s - cached", user["Uuid"])


def orgunit_uuids():
    return os2sync_get("{BASE}/orgunit").json()


def delete_orgunit(uuid):
    if not already_xferred("/orgUnit/" + uuid, {}, "delete"):
        logger.info("delete orgunit %s", uuid)
        os2sync_delete("{BASE}/orgUnit/" + uuid)
    else:
        logger.info("delete orgunit %s - cached", uuid)


def upsert_orgunit(org_unit):
    if not already_xferred("/orgUnit/" + org_unit["Uuid"], org_unit, "upsert"):
        logger.info("upsert orgunit %s", org_unit["Uuid"])
        os2sync_post("{BASE}/orgUnit/", json=org_unit)
    else:
        logger.info("upsert orgunit %s - cached", org_unit["Uuid"])



import math

from integrations.calculate_primary.common import logger
from integrations.calculate_primary.common import MOPrimaryEngagementUpdater


class OPUSPrimaryEngagementUpdater(MOPrimaryEngagementUpdater):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Currently primary is set first by engagement type (order given in
        # settings) and secondly by job_id.
        # TODO: Check that configured eng_types exist
        self.eng_types_order = self.settings[
            "integrations.opus.eng_types_primary_order"
        ]

        def engagements_included_in_primary_calculation(user_uuid, no_past, engagement):
            if engagement["org_unit"]["uuid"] in self.settings.get(
                "integrations.ad.import_ou.mo_unit_uuid", ""
            ):
                # disregard engagements from externals
                logger.warning(
                    "disregarding external engagement: {}".format(engagement)
                )
                return False
            return True

        def remove_missing_user_key(user_uuid, no_past, engagement):
            return "user_key" in engagement

        self.calculate_filters = [
            engagements_included_in_primary_calculation,
            remove_missing_user_key,
        ]

    def _find_primary_types(self):
        """
        Read the engagement types from MO and match them up against the three
        known types in the OPUS->MO import.
        :param helper: An instance of mora-helpers.
        :return: A dict matching up the engagement types with LoRa class uuids.
        """
        # These constants are global in all OPUS municipalities (because they are
        # created by the OPUS->MO importer.
        PRIMARY = "primary"
        NON_PRIMARY = "non-primary"
        FIXED_PRIMARY = "explicitly-primary"

        logger.info("Read primary types")
        primary_dict = {"fixed_primary": None, "primary": None, "non_primary": None}

        primary_types = self.helper.read_classes_in_facet("primary_type")
        for primary_type in primary_types[0]:
            if primary_type["user_key"] == PRIMARY:
                primary_dict["primary"] = primary_type["uuid"]
            if primary_type["user_key"] == NON_PRIMARY:
                primary_dict["non_primary"] = primary_type["uuid"]
            if primary_type["user_key"] == FIXED_PRIMARY:
                primary_dict["fixed_primary"] = primary_type["uuid"]

        if None in primary_dict.values():
            raise Exception("Missing primary types: {}".format(primary_dict))
        primary_list = [primary_dict["fixed_primary"], primary_dict["primary"]]

        return primary_dict, primary_list

    def _find_primary(self, mo_engagements):
        # The primary engagement is the engagement with the lowest engagement type.
        # - The order of engagement types is given by self.eng_types_order.
        #
        # If two engagements have the same engagement_type, the tie is broken by
        # picking the one with the lowest user-key integer.
        def get_engagement_type_id(engagement):
            if engagement["engagement_type"]["uuid"] in self.eng_types_order:
                return self.eng_types_order.index(engagement["engagement_type"]["uuid"])
            return math.inf

        def get_engagement_order(engagement):
            try:
                eng_id = int(engagement["user_key"])
                return eng_id
            except Exception as exp:
                logger.warning(
                    "Skippning engangement with non-integer employment_id: {}".format(
                        engagement["user_key"]
                    )
                )
                logger.exception(exp)
                return math.inf

        primary_engagement = min(
            mo_engagements,
            # Sort first by engagement_type, then by user_key integer
            key=lambda eng: (get_engagement_type_id(eng), get_engagement_order(eng)),
        )
        return primary_engagement["uuid"]



# -*- coding: utf-8 -*-
import json
import logging
import random
import re
import time
from abc import ABC
from abc import abstractmethod
from datetime import date
from datetime import datetime
from functools import partial
from operator import itemgetter

import click
from click_option_group import optgroup
from click_option_group import RequiredMutuallyExclusiveOptionGroup
from jinja2 import Template
from more_itertools import unzip
from os2mo_helpers.mora_helpers import MoraHelper
from ra_utils.lazy_dict import LazyDict
from ra_utils.lazy_dict import LazyEval
from ra_utils.lazy_dict import LazyEvalDerived

from . import ad_templates
from .ad_common import AD
from .ad_exceptions import CprNotFoundInADException
from .ad_exceptions import CprNotNotUnique
from .ad_exceptions import EngagementDatesError
from .ad_exceptions import NoActiveEngagementsException
from .ad_exceptions import NoPrimaryEngagementException
from .ad_exceptions import ReplicationFailedException
from .ad_exceptions import SamAccountNameNotUnique
from .ad_exceptions import UserNotFoundException
from .ad_logger import start_logging
from .ad_template_engine import prepare_field_templates
from .ad_template_engine import template_powershell
from .user_names import UserNameGen
from .utils import dict_exclude
from .utils import dict_map
from .utils import dict_subset
from .utils import lower_list


logger = logging.getLogger("AdWriter")


class MODataSource(ABC):
    @abstractmethod
    def read_user(self, uuid):
        """Read a user from MO using the provided uuid.

        Throws UserNotFoundException if the user cannot be found.

        Args:
            uuid: UUID for the user to lookup.

        Returns:
            dict: A dict with the users data.
        """
        raise NotImplementedError

    @abstractmethod
    def get_email_address(self, uuid):
        """Read a users email address using the provided uuid.

        Args:
            uuid: UUID for the user to lookup.

        Returns:
            dict: A dict of email address, potentially empty.
        """
        raise NotImplementedError

    @abstractmethod
    def find_primary_engagement(self, uuid):
        """Find the primary engagement for the provided uuid user.

        Args:
            uuid: UUID for the user to find primary engagement for.

        Returns:
            tuple(string, string, string, string):
                employment_number: Identifier for the engagement.
                title: Title of the job function for the engagement
                eng_org_unit: UUID of the organisation unit for the engagement
                eng_uuid: UUID of the found engagement
        """
        raise NotImplementedError

    @abstractmethod
    def get_manager_uuid(self, mo_user, eng_uuid):
        """Get UUID of the relevant manager for the user.

        Args:
            mo_user: MO user object, as returned by read_user.
            eng_uuid: UUID of the engagement, as returned by find_primary_engagement.

        Returns:
            str: A UUID string for the manager
        """
        raise NotImplementedError

    @abstractmethod
    def get_engagement_dates(self, uuid):
        """Return all present and future engagement start dates and end dates.

        Args:
            uuid: UUID for the user to lookup.

        Returns:
            tuple[list[str], list[str]]: A tuple of lists of start and end dates.
        """
        raise NotImplementedError

    def get_engagement_endpoint_dates(self, uuid):
        """Return the earliest start- and latest end-date for the users engagements.

        Args:
            uuid: UUID for the user to lookup.

        Returns:
            tuple[str, str]: A tuple with start and end date.
        """
        start_dates, end_dates = self.get_engagement_dates(uuid)

        start_dates = map(lambda date: date if date else "1930-01-01", start_dates)
        start_date = min(start_dates, default="9999-12-31")

        end_dates = map(lambda date: date if date else "9999-12-31", end_dates)
        end_date = max(end_dates, default="1930-01-01")

        if date.fromisoformat(start_date) > date.fromisoformat(end_date):
            raise EngagementDatesError(
                "Invalid Engagement dates interval. "
                f"Start date {start_date} is greater than end date {end_date}"
            )

        return start_date, end_date

    @abstractmethod
    def get_it_systems(self, uuid):
        """Read the IT system bindings from the user using the provided uuid.

        Args:
            uuid: UUID for the user to lookup.

        Returns:
            dict of dicts: A dictionary it-system uuid to it-system dictionaries.
        """
        raise NotImplementedError


class LoraCacheSource(MODataSource):
    """LoraCache implementation of the MODataSource interface."""

    def __init__(self, lc, lc_historic, mo_rest_source):
        self.lc = lc
        self.lc_historic = lc_historic
        self.mo_rest_source = mo_rest_source

    def read_user(self, uuid):
        if uuid not in self.lc.users:
            raise UserNotFoundException()

        lc_user = self.lc.users[uuid][0]
        mo_user = {
            "uuid": uuid,
            "name": lc_user["navn"],
            "surname": lc_user["efternavn"],
            "givenname": lc_user["fornavn"],
            "nickname": lc_user["kaldenavn"],
            "nickname_givenname": lc_user["kaldenavn_fornavn"],
            "nickname_surname": lc_user["kaldenavn_efternavn"],
            "cpr_no": lc_user["cpr"],
        }
        return mo_user

    def get_email_address(self, uuid):
        mail_dict = {}
        for addr in self.lc.addresses.values():
            if addr[0]["user"] == uuid and addr[0]["scope"] == "E-mail":
                mail_dict = addr[0]
        return dict_subset(mail_dict, ["uuid", "value"])

    def find_primary_engagement(self, uuid):
        def filter_for_user(engagements):
            return filter(lambda eng: eng[0]["user"] == uuid, engagements)

        def filter_primary(engagements):
            return filter(lambda eng: eng[0]["primary_boolean"], engagements)

        user_engagements = list(filter_for_user(self.lc.engagements.values()))
        # No user engagements
        if not user_engagements:
            # But we may still have future engagements
            future_engagement = next(
                filter_for_user(self.lc_historic.engagements.values()), None
            )
            # We do not have any engagements at all
            if future_engagement is None:
                raise NoActiveEngagementsException()
            # We have future engagements, but LoraCache does not handle that.
            # Delegate to MORESTSource
            logger.info("Found future engagement")
            return self.mo_rest_source.find_primary_engagement(uuid)

        primary_engagement = next(filter_primary(user_engagements), None)
        if primary_engagement is None:
            raise NoPrimaryEngagementException("User: {}".format(uuid))

        primary_engagement = primary_engagement[0]
        employment_number = primary_engagement["user_key"]
        title = self.lc.classes[primary_engagement["job_function"]]["title"]
        eng_org_unit = primary_engagement["unit"]
        eng_uuid = primary_engagement["uuid"]
        return employment_number, title, eng_org_unit, eng_uuid

    def get_manager_uuid(self, mo_user, eng_uuid):
        def org_uuid_parent(org_uuid):
            parent_uuid = self.lc.units[org_uuid][0]["parent"]
            return parent_uuid

        def org_uuid_to_manager(org_uuid):
            org_unit = self.lc.units[org_uuid][0]
            manager_uuid = self.lc.managers[org_unit["acting_manager_uuid"]][0]["user"]
            return manager_uuid

        try:
            # Compatibility to mimic MORESTSource behaviour
            # MORESTSource does an engagement lookup in the present, using
            # the org uuid from that and fails if it doesn't find anything
            engagement = self.lc.engagements[eng_uuid][0]
            eng_org_unit = engagement["unit"]
            manager_uuid = org_uuid_to_manager(eng_org_unit)
            if manager_uuid is None:
                raise Exception("Unable to find manager")
            # We found a manager directly
            if manager_uuid != mo_user["uuid"]:
                return manager_uuid
            # Self manager, find a manager above us, if possible
            parent_uuid = org_uuid_parent(eng_org_unit)
            while manager_uuid == mo_user["uuid"]:
                if parent_uuid is None:
                    return manager_uuid
                manager_uuid = org_uuid_to_manager(parent_uuid)
                parent_uuid = org_uuid_parent(parent_uuid)
            return manager_uuid
        except KeyError:
            return None

    def get_engagement_dates(self, uuid):
        """Return all present and future engagement start dates and end dates.

        Args:
            uuid: UUID for the user to lookup.

        Returns:
            tuple[list[str], list[str]]: A tuple of lists of start and end dates.
        """
        return self.mo_rest_source.get_engagement_dates(uuid)

    def get_it_systems(self, uuid):
        user_itsystems = filter(
            lambda eng: eng["user"] == uuid,
            map(itemgetter(0), self.lc.it_connections.values()),
        )
        return {it_system["itsystem"]: it_system for it_system in user_itsystems}


class MORESTSource(MODataSource):
    """MO REST implementation of the MODataSource interface."""

    def __init__(self, settings):
        self.helper = MoraHelper(
            hostname=settings["global"]["mora.base"], use_cache=False
        )

    def read_user(self, uuid):
        mo_user = self.helper.read_user(user_uuid=uuid)
        if "uuid" not in mo_user:
            raise UserNotFoundException()
        else:
            assert mo_user["uuid"] == uuid
        exclude_fields = ["org", "user_key"]
        mo_user = dict_exclude(mo_user, exclude_fields)
        return mo_user

    def get_email_address(self, uuid):
        mail_dict = self.helper.get_e_address(uuid, scope="EMAIL")
        return dict_subset(mail_dict, ["uuid", "value"])

    def find_primary_engagement(self, uuid):
        def filter_primary(engagements):
            return filter(lambda eng: eng["is_primary"], engagements)

        user_engagements = self.helper.read_user_engagement(
            uuid, calculate_primary=True, read_all=True, skip_past=True
        )
        if not user_engagements:
            raise NoActiveEngagementsException()

        primary_engagement = next(filter_primary(user_engagements), None)
        if primary_engagement is None:
            raise NoPrimaryEngagementException("User: {}".format(uuid))

        employment_number = primary_engagement["user_key"]
        title = primary_engagement["job_function"]["name"]
        eng_org_unit = primary_engagement["org_unit"]["uuid"]
        eng_uuid = primary_engagement["uuid"]
        return employment_number, title, eng_org_unit, eng_uuid

    def get_manager_uuid(self, mo_user, eng_uuid):
        try:
            manager = self.helper.read_engagement_manager(eng_uuid)
            manager_uuid = manager["uuid"]
            return manager_uuid
        except KeyError:
            return None

    def get_engagement_dates(self, uuid):
        """Return all present and future engagement start dates and end dates.

        Args:
            uuid: UUID for the user to lookup.

        Returns:
            tuple[list[str], list[str]]: A tuple of lists of start and end dates.
        """
        user_engagements = self.helper.read_user_engagement(
            uuid, read_all=True, skip_past=True
        )
        dates = map(itemgetter("validity"), user_engagements)
        dates = map(itemgetter("from", "to"), dates)
        unzipped = unzip(dates)
        if len(unzipped) == 0:
            return [], []
        from_dates, to_dates = unzipped
        return from_dates, to_dates

    def get_it_systems(self, uuid):
        itsystems = self.helper.get_e_itsystems(uuid)

        def to_lora_itsystem(it_system):
            print(it_system)
            return it_system["itsystem"]["uuid"], {
                "uuid": it_system["uuid"],
                "user": it_system["person"]["uuid"],
                "unit": (it_system.get("org_unit") or {"uuid": None}).get("uuid"),
                "username": it_system["user_key"],
                "itsystem": it_system["itsystem"]["uuid"],
                "from_date": it_system["validity"]["from"],
                "to_date": it_system["validity"]["to"],
            }

        return dict(map(to_lora_itsystem, itsystems))


class ADWriter(AD):
    def __init__(self, lc=None, lc_historic=None, **kwargs):
        super().__init__(**kwargs)
        self.settings = self.all_settings
        self.skip_occupied_names = kwargs.get("skip_occupied_names", False)

        # Setup datasource for getting MO data.
        # TODO: Create a factory instead of this hackery?
        # Default to using MORESTSource as data source
        self.datasource = MORESTSource(self.settings)
        # Use LoraCacheSource if LoraCache is provided
        if lc and lc_historic:
            self.datasource = LoraCacheSource(lc, lc_historic, self.datasource)
        # NOTE: These should be eliminated when all uses are gone
        # NOTE: Once fully utilized, tests should be able to just implement a
        #       MODataSource for all their mocking needs.
        self.lc = lc
        self.lc_historic = lc_historic
        self.helper = MoraHelper(
            hostname=self.settings["global"]["mora.base"], use_cache=False
        )

        self._init_name_creator()

    def _init_name_creator(self):
        self.name_creator = UserNameGen.get_implementation()
        if not self.skip_occupied_names:
            logger.info("Reading occupied names")
            self.name_creator.load_occupied_names()
        logger.info("Done reading occupied names")

    def _get_write_setting(self):
        if not self.all_settings["primary_write"]:
            msg = "Trying to enable write access with broken settings."
            logger.error(msg)
            raise Exception(msg)
        return self.all_settings["primary_write"]

    def _wait_for_replication(self, sam):
        t_start = time.time()
        logger.debug("Wait for replication of {}".format(sam))
        if not self.all_settings["global"]["servers"]:
            logger.info("No server infomation, falling back to waiting")
            time.sleep(15)
        else:
            # TODO, read from all AD servers and see when user is available
            replication_finished = False
            while not replication_finished:
                if time.time() - t_start > 60:
                    logger.error("Replication error")
                    raise ReplicationFailedException()

                for server in self.all_settings["global"]["servers"]:
                    user = self.get_from_ad(user=sam, server=server)
                    logger.debug("Testing {}, found: {}".format(server, len(user)))
                    if user:
                        logger.debug("Found successfully")
                        replication_finished = True
                    else:
                        logger.debug("Did not find")
                        replication_finished = False
                        time.sleep(0.25)
                        break
        logger.info("replication_finished: {}s".format(time.time() - t_start))

    def _read_user(self, uuid):
        return self.datasource.read_user(uuid)

    def _find_unit_info(self, eng_org_unit):
        # TODO: Convert to datasource
        write_settings = self._get_write_setting()

        level2orgunit = "Ingen"
        unit_info = {}
        if self.lc:
            unit_name = self.lc.units[eng_org_unit][0]["name"]
            unit_user_key = self.lc.units[eng_org_unit][0]["user_key"]
            location = self.lc.units[eng_org_unit][0]["location"]

            # We initialize parent as the unit itself to ensure to catch if
            # a person is engaged directly in a level2org
            parent_uuid = self.lc.units[eng_org_unit][0]["uuid"]
            while parent_uuid is not None:
                parent_unit = self.lc.units[parent_uuid][0]
                if write_settings["level2orgunit_type"] in (
                    parent_unit["unit_type"],
                    parent_unit["level"],
                ):
                    level2orgunit = parent_unit["name"]
                parent_uuid = parent_unit["parent"]
        else:
            mo_unit_info = self.helper.read_ou(eng_org_unit)
            unit_name = mo_unit_info["name"]
            unit_user_key = mo_unit_info["user_key"]

            location = ""
            current_unit = mo_unit_info
            while current_unit:
                location = current_unit["name"] + "\\" + location
                current_type = current_unit["org_unit_type"]
                current_level = current_unit["org_unit_level"]
                if current_level is None:
                    current_level = {"uuid": None}
                if write_settings["level2orgunit_type"] in (
                    current_type["uuid"],
                    current_level["uuid"],
                ):
                    level2orgunit = current_unit["name"]
                current_unit = current_unit["parent"]
            location = location[:-1]

        unit_info = {
            "name": unit_name,
            "user_key": unit_user_key,
            "location": location,
            "level2orgunit": level2orgunit,
        }
        return unit_info

    def _read_user_addresses(self, eng_org_unit):
        # TODO: Convert to datasource
        addresses = {}
        if self.lc:
            email = []
            postal = {}
            for addr in self.lc.addresses.values():
                if addr[0]["unit"] == eng_org_unit:
                    if addr[0]["scope"] == "DAR":
                        postal = {"Adresse": addr[0]["value"]}
                    if addr[0]["scope"] == "E-mail":
                        visibility = addr[0]["visibility"]
                        visibility_class = None
                        if visibility is not None:
                            visibility_class = self.lc.classes[visibility]
                        email.append(
                            {"visibility": visibility_class, "value": addr[0]["value"]}
                        )
        else:
            email = self.helper.read_ou_address(
                eng_org_unit, scope="EMAIL", return_all=True
            )
            postal = self.helper.read_ou_address(
                eng_org_unit, scope="DAR", return_all=False
            )

        unit_secure_email = None
        unit_public_email = None
        for mail in email:
            if mail["visibility"] is None:
                # If visibility is not set, we assume it is non-public.
                unit_secure_email = mail["value"]
            else:
                if mail["visibility"]["scope"] == "PUBLIC":
                    unit_public_email = mail["value"]
                if mail["visibility"]["scope"] == "SECRET":
                    unit_secure_email = mail["value"]

        addresses = {
            "unit_secure_email": unit_secure_email,
            "unit_public_email": unit_public_email,
            "postal": postal,
        }
        return addresses

    def read_ad_information_from_mo(self, uuid, read_manager=True, ad_dump=None):
        """
        Retrive the necessary information from MO to contruct a new AD user.
        The final information object should of this type, notice that end-date
        is not necessarily for the current primary engagement, but the end-date
        of the longest running currently known primary engagement:
        mo_values = {
            'name': ('Martin Lee', 'Gore'),
            'employment_number': '101',
            'uuid': '7ccbd9aa-gd60-4fa1-4571-0e6f41f6ebc0',
            'end_date': 2089-11-11,
            'cpr': '1122334455',
            'title': 'Musiker',
            'location': 'Viborg Kommune\\Forvalting\\Enhed\\',
            'level2orgunit: 'Beskæftigelse, Økonomi & Personale',
            'manager_sam': 'DMILL'
        }
        """
        logger.info("Read information for {}".format(uuid))
        try:
            (
                employment_number,
                title,
                eng_org_unit,
                eng_uuid,
            ) = self.datasource.find_primary_engagement(uuid)
        except NoActiveEngagementsException:
            logger.info("No active engagements found")
            return None

        def split_addresses(addresses):
            postal_code = city = streetname = "Ukendt"
            if addresses.get("postal"):
                postal = addresses["postal"]
                try:
                    postal_code = re.findall("[0-9]{4}", postal["Adresse"])[0]
                    city_pos = postal["Adresse"].find(postal_code) + 5
                    city = postal["Adresse"][city_pos:]
                    streetname = postal["Adresse"][: city_pos - 7]
                except IndexError:
                    logger.error("Unable to read adresse from MO (no access to DAR?)")
                except TypeError:
                    logger.error("Unable to read adresse from MO (no access to DAR?)")
            return {
                "postal_code": postal_code,
                "city": city,
                "streetname": streetname,
            }

        def read_manager_uuid(mo_user, eng_uuid):
            manager_uuid = self.datasource.get_manager_uuid(mo_user, eng_uuid)
            if manager_uuid is None:
                logger.info("No managers found")
            return manager_uuid

        def read_manager_mail(manager_uuid):
            manager_mail_dict = self.datasource.get_email_address(manager_uuid)
            if manager_mail_dict:
                return manager_mail_dict["value"]
            return None

        def read_manager_sam(manager_cpr):
            try:
                manager_ad_user = self._find_ad_user(cpr=manager_cpr, ad_dump=ad_dump)
            except CprNotFoundInADException:
                logger.info("manager not found by cpr lookup")
            except CprNotNotUnique:
                logger.info("multiple managers found by cpr lookup")
            else:
                return self._get_sam_for_ad_user(manager_ad_user)
            return None

        # NOTE: Underscore fields should not be read
        mo_values: LazyDict = LazyDict(
            {
                # Raw information
                "uuid": uuid,
                # Engagement information
                "employment_number": employment_number,
                "title": title,
                "unit_uuid": eng_org_unit,
                "_eng_uuid": eng_uuid,
                "_dates": LazyEvalDerived(
                    lambda uuid: self.datasource.get_engagement_endpoint_dates(uuid)
                ),
                "start_date": LazyEvalDerived(lambda _dates: _dates[0]),
                "end_date": LazyEvalDerived(lambda _dates: _dates[1]),
                # Lazy MO User and associated fields
                "_mo_user": LazyEvalDerived(lambda uuid: self._read_user(uuid)),
                "name": LazyEvalDerived(
                    lambda _mo_user: (_mo_user["givenname"], _mo_user["surname"])
                ),
                "full_name": LazyEvalDerived(
                    lambda name: "{} {}".format(*name).strip()
                ),
                "nickname": LazyEvalDerived(
                    lambda _mo_user: (
                        _mo_user["nickname_givenname"],
                        _mo_user["nickname_surname"],
                    )
                ),
                "full_nickname": LazyEvalDerived(
                    lambda nickname: "{} {}".format(*nickname).strip()
                ),
                "cpr": LazyEvalDerived(lambda _mo_user: _mo_user["cpr_no"]),
                # Lazy Unit and associated fields
                "_unit": LazyEvalDerived(
                    lambda unit_uuid: self._find_unit_info(unit_uuid)
                ),
                "unit": LazyEvalDerived(lambda _unit: _unit["name"]),
                "unit_user_key": LazyEvalDerived(lambda _unit: _unit["user_key"]),
                "location": LazyEvalDerived(lambda _unit: _unit["location"]),
                "level2orgunit": LazyEvalDerived(lambda _unit: _unit["level2orgunit"]),
                # Lazy addresses and associated fields
                "_addresses": LazyEvalDerived(
                    lambda unit_uuid: self._read_user_addresses(unit_uuid)
                ),
                "_parsed_addresses": LazyEvalDerived(
                    lambda _addresses: split_addresses(_addresses)
                ),
                "unit_postal_code": LazyEvalDerived(
                    lambda _parsed_addresses: _parsed_addresses["postal_code"]
                ),
                "unit_city": LazyEvalDerived(
                    lambda _parsed_addresses: _parsed_addresses["city"]
                ),
                "unit_streetname": LazyEvalDerived(
                    lambda _parsed_addresses: _parsed_addresses["streetname"]
                ),
                "unit_public_email": LazyEvalDerived(
                    lambda _addresses: _addresses["unit_public_email"]
                ),
                "unit_secure_email": LazyEvalDerived(
                    lambda _addresses: _addresses["unit_secure_email"]
                ),
                # Manager stuff
                "_manager_uuid": LazyEval(
                    lambda key, dictionary: (
                        read_manager_uuid(
                            dictionary["_mo_user"], dictionary["_eng_uuid"]
                        )
                        if read_manager
                        else None
                    )
                ),
                "_manager_mo_user": LazyEvalDerived(
                    lambda _manager_uuid: self._read_user(_manager_uuid)
                    if _manager_uuid
                    else {}
                ),
                "manager_name": LazyEvalDerived(
                    lambda _manager_mo_user: _manager_mo_user.get("name")
                ),
                "manager_cpr": LazyEvalDerived(
                    lambda _manager_mo_user: _manager_mo_user.get("cpr_no")
                ),
                "manager_mail": LazyEvalDerived(
                    lambda _manager_uuid: read_manager_mail(_manager_uuid)
                    if _manager_uuid
                    else None
                ),
                "manager_sam": LazyEvalDerived(
                    lambda manager_cpr: read_manager_sam(manager_cpr)
                    if manager_cpr
                    else None
                ),
                "read_manager": LazyEvalDerived(
                    lambda _manager_uuid: bool(_manager_uuid)
                ),
                # IT systems
                "it_systems": LazyEvalDerived(
                    lambda uuid: self.datasource.get_it_systems(uuid)
                ),
            }
        )
        return mo_values

    def add_manager_to_user(self, user_sam, manager_sam):
        """
        Mark an existing AD user as manager for an existing AD user.
        :param user_sam: SamAccountName for the employee.
        :param manager_sam: SamAccountName for the manager.
        """
        format_rules = {"user_sam": user_sam, "manager_sam": manager_sam}
        ps_script = self._build_ps(ad_templates.add_manager_template, format_rules)

        response = self._run_ps_script(ps_script)
        return response is {}

    def _rename_ad_user(self, user_sam, new_name):
        logger.info("Rename user: %s", user_sam)
        # Todo: This code is a duplicate of code found elsewhere
        rename_user_template = ad_templates.rename_user_template
        rename_user_string = rename_user_template.format(
            user_sam=user_sam,
            new_name=new_name,
        )
        rename_user_string = self.remove_redundant(rename_user_string)
        server_string = ""
        if self.all_settings["global"].get("servers") is not None:
            server_string = " -Server {} ".format(
                random.choice(self.all_settings["global"]["servers"])
            )
        ps_script = self._build_user_credential() + rename_user_string + server_string
        logger.debug("Rename user, ps_script: {}".format(ps_script))
        response = self._run_ps_script(ps_script)
        logger.debug("Response from sync: {}".format(response))
        logger.debug("Wait for replication")
        # Todo: In principle we should ask all DCs, bu this will happen
        # very rarely, performance is not of great importance
        time.sleep(10)

    def _cf(self, ad_field, value, ad):
        logger.info("Check AD field: {}".format(ad_field))
        mismatch = {}
        if value is None:
            msg = "Value for {} is None-type replace to string None"
            logger.debug(msg.format(ad_field))
            value = "None"
        if ad.get(ad_field) != value:
            msg = "{}: AD value: {}, does not match MO value: {}"
            logger.info(msg.format(ad_field, ad.get(ad_field), value))
            mismatch = {ad_field: (ad.get(ad_field), value)}
        return mismatch

    def _render_field_template(self, context, template):
        return Template(template.strip('"')).render(**context)

    def _sync_compare(self, mo_values, ad_dump):
        ad_user = self._find_ad_user(mo_values["cpr"], ad_dump=ad_dump)
        user_sam = self._get_sam_for_ad_user(ad_user)

        # TODO: Why is this not generated along with all other info in mo_values?
        mo_values["name_sam"] = "{} - {}".format(mo_values["full_name"], user_sam)

        fields = prepare_field_templates("Set-ADUser", settings=self.all_settings)

        def to_lower(string):
            return string.lower()

        ad_user = dict_map(ad_user, key_func=to_lower)
        fields = dict_map(fields, key_func=to_lower)

        never_compare = lower_list(["Credential", "Manager"])
        fields = dict_exclude(fields, never_compare)

        context = {
            "ad_values": ad_user,
            "mo_values": mo_values,
            "user_sam": user_sam,
        }

        # Build context and render template to get comparision value
        # NOTE: This results in rendering the template twice, once here and
        #       once inside the powershell render call.
        #       We should probably restructure this, such that we only render
        #       the template once, potentially rendering a dict of results.
        # TODO: Make the above mentioned change.
        fields = dict_map(
            fields,
            value_func=partial(self._render_field_template, context),
        )
        mismatch = {}
        for ad_field, rendered_value in fields.items():
            mismatch.update(self._cf(ad_field, rendered_value, ad_user))

        if mo_values.get("manager_cpr"):
            manager_ad_user = self._find_ad_user(
                mo_values["manager_cpr"], ad_dump=ad_dump
            )
            manager_distinguished_name = manager_ad_user["DistinguishedName"]
            if ad_user["manager"] != manager_distinguished_name:
                mismatch["manager"] = (ad_user["manager"], manager_distinguished_name)
                logger.info("Manager should be updated")

        return mismatch

    def sync_user(self, mo_uuid, ad_dump=None, sync_manager=True):
        """
        Sync MO information into AD
        """
        mo_values = self.read_ad_information_from_mo(
            mo_uuid, ad_dump=ad_dump, read_manager=sync_manager
        )

        if mo_values is None:
            return (False, "No active engagments")

        ad_user = self._find_ad_user(mo_values["cpr"], ad_dump=ad_dump)
        user_sam = self._get_sam_for_ad_user(ad_user)

        if ad_dump is None:
            # TODO: We could also add the compare logic here,
            # but the benefit will be max 40%
            mismatch = {"force re-sync": "yes", "manager": "yes"}
        else:
            mismatch = self._sync_compare(mo_values, ad_dump)

        logger.debug("Sync compare: {}".format(mismatch))

        if "name" in mismatch:
            response = self._rename_ad_user(user_sam, mismatch["name"][1])
            del mismatch["name"]

        if not mismatch and ("sync_timestamp" not in str(self.all_settings)):
            # If "sync_timestamp" is in settings we assume the intent is to always write a timestamp.
            logger.info("Nothing to edit")
            return (True, "Nothing to edit", mo_values["read_manager"])

        logger.info("Sync compare: {}".format(mismatch))

        edit_user_string = template_powershell(
            cmd="Set-ADUser",
            context={
                "ad_values": ad_user,
                "mo_values": mo_values,
                "user_sam": user_sam,
                "sync_timestamp": str(datetime.now()),
            },
            settings=self.all_settings,
        )
        edit_user_string = self.remove_redundant(edit_user_string)

        server_string = ""
        if self.all_settings["global"].get("servers") is not None:
            server_string = " -Server {} ".format(
                random.choice(self.all_settings["global"]["servers"])
            )

        ps_script = self._build_user_credential() + edit_user_string + server_string
        logger.debug("Sync user, ps_script: {}".format(ps_script))

        response = self._run_ps_script(ps_script)
        logger.debug("Response from sync: {}".format(response))

        if sync_manager and "manager" in mismatch:
            logger.info("Add manager")
            self.add_manager_to_user(
                user_sam=user_sam, manager_sam=mo_values["manager_sam"]
            )

        return (True, "Sync completed", mo_values["read_manager"])

    def create_user(self, mo_uuid, create_manager, dry_run=False):
        """
        Create an AD user
        :param mo_uuid: uuid for the MO user we want to add to AD.
        :param create_manager: If True, an AD link will be added between the user
        object and the AD object of the users manager.
        :param dry_run: Not yet implemented. Should return whether the user is
        expected to be able to be created in AD and the expected SamAccountName.
        :return: The generated SamAccountName for the new user
        """
        # TODO: Implement dry_run

        bp = self._ps_boiler_plate()
        mo_values = self.read_ad_information_from_mo(mo_uuid, create_manager)
        if mo_values is None:
            logger.error("Trying to create user with no engagements")
            raise NoPrimaryEngagementException

        all_names = mo_values["name"][0].split(" ") + [mo_values["name"][1]]
        sam_account_name = self.name_creator.create_username(all_names, dry_run=dry_run)

        existing_sam = self.get_from_ad(user=sam_account_name)
        existing_cpr = self.get_from_ad(cpr=mo_values["cpr"])
        if existing_sam:
            logger.error("SamAccount already in use: {}".format(sam_account_name))
            raise SamAccountNameNotUnique(sam_account_name)
        if existing_cpr:
            logger.error("cpr already in use: {}".format(mo_values["cpr"]))
            raise CprNotNotUnique(mo_values["cpr"])

        create_user_string = template_powershell(
            context={
                "ad_values": {},
                "mo_values": mo_values,
                "user_sam": sam_account_name,
                "sync_timestamp": str(datetime.now()),
            },
            settings=self.all_settings,
        )
        create_user_string = self.remove_redundant(create_user_string)

        # Should this go to self._ps_boiler_plate()?
        server_string = ""
        if self.all_settings["global"].get("servers"):
            server_string = " -Server {} ".format(
                random.choice(self.all_settings["global"]["servers"])
            )

        ps_script = (
            self._build_user_credential()
            + create_user_string
            + server_string
            + bp["path"]
        )

        response = self._run_ps_script(ps_script)
        if not response == {}:
            msg = "Create user failed, message: {}".format(response)
            logger.error(msg)
            return (False, msg)

        if create_manager:
            self._wait_for_replication(sam_account_name)
            print(
                "Add {} as manager for {}".format(
                    mo_values["manager_sam"], sam_account_name
                )
            )
            logger.info(
                "Add {} as manager for {}".format(
                    mo_values["manager_sam"], sam_account_name
                )
            )
            self.add_manager_to_user(
                user_sam=sam_account_name, manager_sam=mo_values["manager_sam"]
            )

        return (True, sam_account_name)

    def add_ad_to_user_it_systems(self, username):
        # TODO: We need a function to write the SamAccount to the user's
        # IT-systems. This is most likely most elegantly done by importing
        # the AD->MO sync tool
        pass

    def set_user_password(self, username, password):
        """
        Set a password for a user.
        :param username: SamAccountName for the user.
        :param password: The password to assign to the user.
        :return: True if success, otherwise False
        """

        format_rules = {"username": username, "password": password}
        ps_script = self._build_ps(ad_templates.set_password_template, format_rules)
        response = self._run_ps_script(ps_script)
        if not response:
            return (True, "Password updated")
        else:
            msg = "Failed to set password!: {}".format(response)
            logger.error(msg)
            return (False, msg)

    def enable_user(self, username, enable=True):
        """
        Enable or disable an AD account.
        :param username: SamAccountName of the account to be enabled or disabled
        :param enable: If True enable account, if False, disbale account
        """

        logger.info("Enable account: {}".format(enable))
        format_rules = {"username": username}
        if enable:
            ps_script = self._build_ps(ad_templates.enable_user_template, format_rules)
        else:
            ps_script = self._build_ps(ad_templates.disable_user_template, format_rules)

        response = self._run_ps_script(ps_script)
        if not response:
            return (True, "Account enabled or disabled")
        else:
            msg = "Failed to update account!: {}".format(response)
            logger.error(msg)
            return (False, msg)

    def delete_user(self, username):
        """
        Delete an AD account. Only to be used for service purpose, actual
        AD integration should never delete a user, but only mark it for
        deletetion.
        :param username: SamAccountName of the account to be deleted
        """
        format_rules = {"username": username}
        ps_script = self._build_ps(
            ad_templates.delete_user_template, format_rules=format_rules
        )

        response = self._run_ps_script(ps_script)
        # TODO: Should we make a read to confirm the user is gone?
        if not response:
            return (True, "User deleted")
        else:
            logger.error("Failed to delete account!: {}".format(response))
            return (False, "Failed to delete")


@click.command()
@optgroup.group("Action", cls=RequiredMutuallyExclusiveOptionGroup)
@optgroup.option(
    "--create-user-with-manager",
    help="Create a new user in AD, also assign a manager",
)
@optgroup.option(
    "--create-user",
    help="Create a new user in AD, do not assign a manager",
)
@optgroup.option(
    "--sync-user",
    help="Sync relevant fields from MO to AD",
)
@optgroup.option(
    "--mo-values",
    type=click.UUID,
    help="Show mo-values for the user",
)
@optgroup.option("--delete-user")
@optgroup.option("--read-ad-information")
@optgroup.option("--add-manager-to-user", nargs=2, type=str)
@click.option("--ignore-occupied-names", is_flag=True, default=False)
def cli(**args):
    """
    Command line interface for the AD writer class.
    """

    ad_writer = ADWriter(skip_occupied_names=args["ignore_occupied_names"])

    if args.get("create_user_with_manager"):
        print("Create_user_with_manager:")
        status = ad_writer.create_user(
            args["create_user_with_manager"], create_manager=True
        )
        # TODO: execute custom script? Or should this be done in
        # two steps.
        print(status[1])

    if args.get("create_user"):
        print("Create user, no link to manager:")
        status = ad_writer.create_user(args["create_user"], create_manager=False)
        print(status[1])

    if args.get("sync_user"):
        print("Sync MO fields to AD")
        status = ad_writer.sync_user(args["sync_user"])
        print(status[1])

    if args.get("delete_user"):
        print("Deleting user:")
        status = ad_writer.delete_user(args["delete_user"])
        print(status[1])

    if args.get("read_ad_information"):
        print("AD information on user:")
        sam = args["read_ad_information"]
        user = ad_writer.get_from_ad(user=sam)
        if not user:
            print("User not found")
        else:
            for key, value in sorted(user[0].items()):
                print("{}: {}".format(key, value))

    if args.get("add_manager_to_user"):
        manager, user = args["add_manager_to_user"]
        print("{} is now set as manager for {}".format(manager, user))
        ad_writer.add_manager_to_user(manager_sam=manager, user_sam=user)

    if args.get("mo_values"):
        mo_values = ad_writer.read_ad_information_from_mo(str(args["mo_values"]))
        print(json.dumps(dict(mo_values.items()), indent=4))


if __name__ == "__main__":
    start_logging("ad_writer.log")
    cli()



from unittest import TestCase
from uuid import uuid4

from ..ad_sync import AdMoSync
from .mocks import MockLoraCache
from .mocks import MockLoraCacheEmptyEmployee
from .mocks import MockMoraHelper


class _TestableAdMoSync(AdMoSync):
    def __init__(self, mock_lora_cache):
        self._mock_lora_cache = mock_lora_cache
        super().__init__(all_settings={})

    def _setup_mora_helper(self):
        return MockMoraHelper("not-a-cpr")

    def _setup_lora_cache(self):
        return self._mock_lora_cache

    def _setup_visibilities(self):
        pass


_mo_values = {"uuid": uuid4()}


class TestAdMoSyncRegressions(TestCase):
    def test_read_all_mo_users_returns_users(self):
        instance = _TestableAdMoSync(MockLoraCache(_mo_values))
        expected_users = [_mo_values]
        actual_users = instance._read_all_mo_users()
        self.assertEqual(actual_users, expected_users)

    def test_read_all_mo_users_crashes_on_empty_user(self):
        """`AdMoSync._read_all_mo_users` raises an `IndexError` if it
        encounters one or more "empty" users in `self.lc.users`.
        (An "empty" user is a key/value pair where the key is a valid UUID but
        the value is an empty list.)
        See #48568.
        """
        instance = _TestableAdMoSync(MockLoraCacheEmptyEmployee(_mo_values))
        with self.assertRaises(IndexError):
            instance._read_all_mo_users()



import os
import sys
import datetime
import pathlib
import json
from os2mo_data_import import ImportHelper
from integrations.SD_Lon import sd_importer

settings = json.loads(pathlib.Path("settings/settings.json").read_text())

importer = ImportHelper(
    create_defaults=True,
    mox_base=settings["mox.base"],
    mora_base=settings["mora.base"],
    system_name='SD-Import',
    end_marker='SDSTOP',
    store_integration_data=True,
    seperate_names=False
)


sd = sd_importer.SdImport(
    importer=importer,
    ad_info=None,
    org_only=True,
)

sd.create_ou_tree(create_orphan_container=True)
importer.import_all()



import sqlite3
from datetime import date
from datetime import datetime
from typing import Any
from typing import cast
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import click
from ra_utils.load_settings import load_setting


class DBOverview:
    def __init__(self, rundb_path: str):
        self.run_db = rundb_path

    def read_db_content(self, *fields):
        conn = sqlite3.connect(self.run_db, detect_types=sqlite3.PARSE_DECLTYPES)
        c = conn.cursor()

        if len(fields) == 0:
            fields = ["*"]

        query = "SELECT " + ", ".join([*fields]) + " FROM runs ORDER BY id"
        c.execute(query)
        rows = c.fetchall()
        for row in rows:
            status = "id: {}, from: {}, to: {}, status: {}"
            print(status.format(*row))

    def _read_last_line(self, *fields) -> Union[Any, List[Any]]:
        conn = sqlite3.connect(self.run_db, detect_types=sqlite3.PARSE_DECLTYPES)
        c = conn.cursor()

        query = "SELECT " + ", ".join([*fields]) + " FROM runs ORDER BY id DESC LIMIT 1"
        c.execute(query)
        result = c.fetchone()
        if len(result) == 1:
            return result[0]
        return result

    def read_current_status(self) -> Tuple[bool, str]:
        today = date.today()
        midnight = datetime.min.time()
        midnight_today = datetime.combine(today, midnight)

        to_date, status = self._read_last_line("to_date", "status")
        if "Running" in status:
            return (False, "Not ready to run")
        if to_date < midnight_today:
            return (False, "Not up to date")
        return (True, "Status ok")

    def _delete_line(self, id: int) -> Tuple[bool, str]:
        conn = sqlite3.connect(self.run_db, detect_types=sqlite3.PARSE_DECLTYPES)
        c = conn.cursor()

        query = "DELETE FROM runs WHERE id=?"
        c.execute(query, (id,))
        conn.commit()
        return (True, "Status ok")

    def delete_last_row(self, force: bool = False) -> Tuple[bool, str]:
        current_status = self.read_current_status()
        if current_status[0] and not force:
            return (True, "Status ok, no delete")

        last_id: int = cast(int, self._read_last_line("id"))
        status, msg = self._delete_line(last_id)
        if status:
            return (True, "Deleted last row")
        return status, msg

    def create_db(self) -> Tuple[bool, str]:
        conn = sqlite3.connect(self.run_db, detect_types=sqlite3.PARSE_DECLTYPES)
        c = conn.cursor()
        c.execute(
            """
            CREATE TABLE runs (
                id INTEGER PRIMARY KEY,
                from_date timestamp,
                to_date timestamp,
                status text
            )
        """
        )
        conn.commit()
        conn.close()
        return (True, "Status ok")


@click.group()
@click.option("--rundb-path", help="Path to the rundb to operate on.")
@click.option("--rundb-variable", help="Settings.json variable to use for rundb path.")
@click.pass_context
def cli(ctx, rundb_path: Optional[str] = None, rundb_variable: Optional[str] = None):
    rundb: str = ""
    if rundb_path:
        rundb = cast(str, rundb_path)
    elif rundb_variable:
        rundb = load_setting(rundb_variable)()
    else:
        raise click.ClickException("Must provide either rundb-path or rundb-variable")

    ctx.ensure_object(dict)
    ctx.obj["dboverview"] = DBOverview(rundb)


@cli.command()
@click.pass_context
def create(ctx):
    """Create a new run_db."""
    db_overview = ctx.obj["dboverview"]
    db_overview.create_db()


@cli.command()
@click.pass_context
def read_rundb(ctx):
    """Load the run_db and print all contents."""
    db_overview = ctx.obj["dboverview"]
    db_overview.read_db_content()


@cli.command()
@click.pass_context
def read_current_status(ctx):
    """Report back the current status of the run_db."""
    db_overview = ctx.obj["dboverview"]
    status, msg = db_overview.read_current_status()
    print(status, msg)

    if not status:
        raise click.ClickException("Job is already running or dates do not match!")


@cli.command()
@click.option(
    "--force",
    is_flag=True,
    default=False,
    help="Remove last line regardless of if it is problematic or not",
)
@click.pass_context
def remove_already_running(ctx, force: bool):
    """Remove problematic entry from the run_db."""
    db_overview = ctx.obj["dboverview"]
    print(db_overview.delete_last_row(force))


if __name__ == "__main__":
    cli()



import json
import logging
import pathlib
from pathlib import Path

import xmltodict
from os2mo_helpers.mora_helpers import MoraHelper
from requests import Session

import constants
from ra_utils.load_settings import load_settings
from integrations import dawa_helper
from integrations.opus import opus_helpers
from integrations.opus.opus_exceptions import UnknownOpusAction

LOG_LEVEL = logging.DEBUG
LOG_FILE = 'mo_integrations.log'

logger = logging.getLogger("opusImport")

for name in logging.root.manager.loggerDict:
    if name in ('opusImport', 'opusHelper', 'moImporterMoraTypes',
                'moImporterMoxTypes', 'moImporterUtilities', 'moImporterHelpers',
                'ADReader'):
        logging.getLogger(name).setLevel(LOG_LEVEL)
    else:
        logging.getLogger(name).setLevel(logging.ERROR)

logging.basicConfig(
    format='%(levelname)s %(asctime)s %(name)s %(message)s',
    level=LOG_LEVEL,
    filename=LOG_FILE
)


class OpusImport(object):

    def __init__(self, importer, org_name, xml_data, ad_reader=None,
                 import_first=False, employee_mapping={}):
        """ If import first is False, the first unit will be skipped """
        self.org_uuid = None

        self.settings = load_settings()
        self.filter_ids = self.settings.get('integrations.opus.units.filter_ids', [])

        self.importer = importer
        self.import_first = import_first
        self.session = Session()
        self.helper = MoraHelper(hostname=self.settings['mora.base'],
                                 use_cache=False)

        self.organisation_id = None
        self.units = None
        self.employees = None
        # Update the above values
        municipality_code = self.parser(xml_data)

        self.org_name = org_name
        self.importer.add_organisation(
            identifier=org_name,
            user_key=org_name,
            municipality_code=municipality_code
        )

        importer.new_itsystem(
            identifier='Opus',
            system_name=constants.Opus_it_system
        )

        self.ad_people = {}
        self.employee_forced_uuids = employee_mapping
        self.ad_reader = None
        if ad_reader:
            self.ad_reader = ad_reader
            self.importer.new_itsystem(
                identifier='AD',
                system_name=constants.AD_it_system
            )
            self.ad_reader.cache_all()

        self.employee_addresses = {}
        self._add_klasse('AddressPostUnit', 'Postadresse',
                         'org_unit_address_type', 'DAR')
        self._add_klasse('Pnummer', 'Pnummer',
                         'org_unit_address_type', 'PNUMBER')
        self._add_klasse('EAN', 'EAN', 'org_unit_address_type', 'EAN')
        self._add_klasse('PhoneUnit', 'Telefon', 'org_unit_address_type', 'PHONE')
        self._add_klasse('PhoneEmployee', 'Telefon', 'employee_address_type',
                         'PHONE')
        self._add_klasse('EmailEmployee', 'Email',
                         'employee_address_type', 'EMAIL')
        self._add_klasse('CVR', 'CVR', 'org_unit_address_type')
        self._add_klasse('SE', 'SE', 'org_unit_address_type')
        self._add_klasse('AdressePostEmployee', 'Postadresse',
                         'employee_address_type', 'DAR')
        self._add_klasse('Lederansvar', 'Lederansvar', 'responsibility')
        self._add_klasse('Ekstern', 'Må vises eksternt', 'visibility', 'PUBLIC')
        self._add_klasse('Intern', 'Må vises internt', 'visibility', 'INTERNAL')
        self._add_klasse('Hemmelig', 'Hemmelig', 'visibility', 'SECRET')

        self._add_klasse('primary', 'Ansat', 'primary_type', '3000')
        self._add_klasse('non-primary', 'Ikke-primær ansættelse',
                         'primary_type', '0')
        self._add_klasse('explicitly-primary', 'Manuelt primær ansættelse',
                         'primary_type', '5000')

    def _update_ad_map(self, cpr):
        logger.debug('Update cpr {}'.format(cpr))
        self.ad_people[cpr] = {}
        if self.ad_reader:
            response = self.ad_reader.read_user(cpr=cpr, cache_only=True)
            if response:
                logger.debug('AD response: {}'.format(response))
                self.ad_people[cpr] = response
            else:
                logger.debug('Not found in AD')

    def insert_org_units(self):
        for unit in self.units:
            self._import_org_unit(unit)

    def insert_employees(self):
        for employee in self.employees:
            self._import_employee(employee)

    def _add_klasse(self, klasse_id, klasse, facet, scope='TEXT'):
        if not self.importer.check_if_exists('klasse', klasse_id):
            uuid = opus_helpers.generate_uuid(klasse_id)
            self.importer.add_klasse(
                identifier=klasse_id,
                uuid=uuid,
                facet_type_ref=facet,
                user_key=klasse_id,
                scope=scope,
                title=klasse
            )
        return klasse_id

    def parser(self, target_file):
        """
        Parse XML data and covert into usable dictionaries

        :return:
        """

        with open(target_file) as xmldump:
            data = xmltodict.parse(xmldump.read())['kmd']

        self.organisation_id = data['orgUnit'][0]['@id']

        if self.import_first:
            self.units = data['orgUnit']
        else:
            self.units = data['orgUnit'][1:]

        self.units = opus_helpers.filter_units(self.units, self.filter_ids)

        self.employees = data['employee']

        municipality_code = int(data['orgUnit'][0]['@client'])
        return municipality_code

    def _import_org_unit(self, unit):
        try:
            org_type = unit['orgType']
            self._add_klasse(org_type, unit['orgTypeTxt'], 'org_unit_type')
        except KeyError:
            org_type = 'Enhed'
            self._add_klasse(org_type, 'Enhed', 'org_unit_type')

        identifier = unit['@id']
        uuid = opus_helpers.generate_uuid(identifier)
        logger.debug('Generated uuid for {}: {}'.format(unit['@id'], uuid))

        user_key = unit['shortName']

        if unit['startDate'] == '1900-01-01':
            date_from = '1930-01-01'
        else:
            date_from = unit['startDate']

        if unit['endDate'] == '9999-12-31':
            date_to = None
        else:
            date_to = unit['endDate']

        name = unit['longName']

        parent_org = unit.get("parentOrgUnit")
        if parent_org == self.organisation_id and not self.import_first:
            parent_org = None

        self.importer.add_organisation_unit(
            identifier=identifier,
            name=name,
            uuid=str(uuid),
            user_key=user_key,
            parent_ref=parent_org,
            type_ref=org_type,
            date_from=date_from,
            date_to=date_to
        )

        if 'seNr' in unit:
            self.importer.add_address_type(
                organisation_unit=identifier,
                value=unit['seNr'],
                type_ref='SE',
                date_from=date_from,
                date_to=date_to
            )

        if 'cvrNr' in unit:
            self.importer.add_address_type(
                organisation_unit=identifier,
                value=unit['cvrNr'],
                type_ref='CVR',
                date_from=date_from,
                date_to=date_to
            )

        if 'eanNr' in unit and (not unit['eanNr'] == '9999999999999'):
            self.importer.add_address_type(
                organisation_unit=identifier,
                value=unit['eanNr'],
                type_ref='EAN',
                date_from=date_from,
                date_to=date_to
            )

        if 'pNr' in unit and (not unit['pNr'] == '0000000000'):
            self.importer.add_address_type(
                organisation_unit=identifier,
                value=unit['pNr'],
                type_ref='Pnummer',
                date_from=date_from,
                date_to=date_to
            )

        if unit['phoneNumber']:
            self.importer.add_address_type(
                organisation_unit=identifier,
                value=unit['phoneNumber'],
                type_ref='PhoneUnit',
                date_from=date_from,
                date_to=date_to
            )

        address_string = unit['street']
        zip_code = unit['zipCode']
        if address_string and zip_code:
            address_uuid = dawa_helper.dawa_lookup(address_string, zip_code)
            if address_uuid:
                self.importer.add_address_type(
                    organisation_unit=identifier,
                    value=address_uuid,
                    type_ref='AddressPostUnit',
                    date_from=date_from,
                    date_to=date_to
                )

    def add_addresses_to_employees(self):
        for cpr, employee_addresses in self.employee_addresses.items():
            for facet, address in employee_addresses.items():
                logger.debug('Add address {}'.format(address))
                if address:
                    self.importer.add_address_type(
                        employee=cpr,
                        value=address,
                        type_ref=facet,
                        date_from='1930-01-01',
                        date_to=None
                    )

    def _import_employee(self, employee):
        # UNUSED KEYS:
        # '@lastChanged'

        logger.debug('Employee object: {}'.format(employee))
        if 'cpr' in employee:
            cpr = employee['cpr']['#text']
            if employee['firstName'] is None and employee['lastName'] is None:
                # Service user, skip
                logger.info('Skipped {}, we think it is a serviceuser'.format(cpr))
                return

        else:  # This employee has left the organisation
            if not employee['@action'] == 'leave':
                msg = 'Unknown action: {}'.format(employee['@action'])
                logger.error(msg)
                raise UnknownOpusAction(msg)
            return

        self._update_ad_map(cpr)

        uuid = self.employee_forced_uuids.get(cpr)
        logger.info('Employee in force list: {} {}'.format(cpr, uuid))
        if uuid is None and 'ObjectGuid' in self.ad_people[cpr]:
            uuid = self.ad_people[cpr]['ObjectGuid']

        date_from = employee['entryDate']
        date_to = employee['leaveDate']
        if date_from is None:
            date_from = '1930-01-01'

        # Only add employee and address information once, this info is duplicated
        # if the employee has multiple engagements
        if not self.importer.check_if_exists('employee', cpr):
            self.employee_addresses[cpr] = {}
            self.importer.add_employee(
                identifier=cpr,
                name=(employee['firstName'], employee['lastName']),
                cpr_no=cpr,
                uuid=uuid
            )

        if 'SamAccountName' in self.ad_people[cpr]:
            self.importer.join_itsystem(
                employee=cpr,
                user_key=self.ad_people[cpr]['SamAccountName'],
                itsystem_ref='AD',
                date_from='1930-01-01'
            )

        if 'userId' in employee:
            self.importer.join_itsystem(
                employee=cpr,
                user_key=employee['userId'],
                itsystem_ref='Opus',
                date_from=date_from,
                date_to=date_to
            )

        if 'email' in employee:
            self.employee_addresses[cpr]['EmailEmployee'] = employee['email']
        if employee['workPhone'] is not None:
            phone = opus_helpers.parse_phone(employee['workPhone'])
            self.employee_addresses[cpr]['PhoneEmployee'] = phone

        if 'postalCode' in employee and employee['address']:
            if isinstance(employee['address'], dict):
                # This is a protected address, cannot import
                pass
            else:
                address_string = employee['address']
                zip_code = employee["postalCode"]
                addr_uuid = dawa_helper.dawa_lookup(address_string, zip_code)
                if addr_uuid:
                    self.employee_addresses[cpr]['AdressePostEmployee'] = addr_uuid

        job = employee["position"]
        self._add_klasse(job, job, 'engagement_job_function')

        if 'workContractText' in employee:
            contract = employee['workContractText']
        else:
            contract = 'Ansat'
        self._add_klasse(contract, contract, 'engagement_type')

        org_unit = employee['orgUnit']
        job_id = employee['@id']
        # engagement_uuid = opus_helpers.generate_uuid(job_id)

        # Every engagement is initially imported as non-primary,
        # a seperate script will correct this after import.
        # This allows separate rules for primary calculation.
        logger.info('Add engagement: {} to {}'.format(job_id, cpr))
        self.importer.add_engagement(
            employee=cpr,
            # uuid=str(engagement_uuid),
            organisation_unit=org_unit,
            primary_ref='non-primary',
            user_key=job_id,
            job_function_ref=job,
            engagement_type_ref=contract,
            date_from=date_from,
            date_to=date_to
        )

        if employee['isManager'] == 'true':
            manager_type_ref = 'manager_type_' + job
            self._add_klasse(manager_type_ref, job, 'manager_type')

            # Opus has two levels of manager_level, since MO handles only one
            # they are concatenated into one.
            manager_level = '{}.{}'.format(employee['superiorLevel'],
                                           employee['subordinateLevel'])
            self._add_klasse(manager_level, manager_level, 'manager_level')
            logger.info('{} is manager {}'.format(cpr, manager_level))
            self.importer.add_manager(
                employee=cpr,
                user_key=job_id,
                organisation_unit=org_unit,
                manager_level_ref=manager_level,
                manager_type_ref=manager_type_ref,
                responsibility_list=['Lederansvar'],
                date_from=date_from,
                date_to=date_to
            )

        if 'function' in employee:
            if not isinstance(employee['function'], list):
                roles = [employee['function']]
            else:
                roles = employee['function']

            for role in roles:
                logger.debug('{} has role {}'.format(cpr, role))
                # We have only a single class for roles, must combine the information
                if 'roleText' in role:
                    combined_role = '{} - {}'.format(role['artText'],
                                                     role['roleText'])
                else:
                    combined_role = role['artText']
                self._add_klasse(combined_role, combined_role, 'role_type')

                date_from = role['@startDate']
                if role['@endDate'] == '9999-12-31':
                    date_to = None
                else:
                    date_to = role['@endDate']

                self.importer.add_role(
                    employee=cpr,
                    organisation_unit=org_unit,
                    role_type_ref=combined_role,
                    date_from=date_from,
                    date_to=date_to
                )

def start_opus_import(importer, ad_reader=None, force=False):
    """
    Start an opus import, run the oldest available dump that
    has not already been imported.
    """
    SETTINGS = load_settings()
    dumps = opus_helpers.read_available_dumps()

    run_db = Path(SETTINGS['integrations.opus.import.run_db'])
    if not run_db.is_file():
        logger.error('Local base not correctly initialized')
        if not force:
            raise RunDBInitException('Local base not correctly initialized')
        else:
            opus_helpers.initialize_db(run_db)
        xml_date = sorted(dumps.keys())[0]
    else:
        if force:
            raise RedundantForceException('Used force on existing db')
        xml_date = opus_helpers.next_xml_file(run_db, dumps)

    xml_file = dumps[xml_date]
    opus_helpers.local_db_insert((xml_date, 'Running since {}'))

    employee_mapping = opus_helpers.read_cpr_mapping()

    opus_importer = OpusImport(
        importer,
        org_name=SETTINGS['municipality.name'],
        xml_data=str(xml_file),
        ad_reader=ad_reader,
        import_first=True,
        employee_mapping=employee_mapping
    )
    logger.info('Start import')
    opus_importer.insert_org_units()
    opus_importer.insert_employees()
    opus_importer.add_addresses_to_employees()
    opus_importer.importer.import_all()
    logger.info('Ended import')

    opus_helpers.local_db_insert((xml_date, 'Import ended: {}'))



#
# Copyright (c) 2021, Magenta ApS
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import logging
from typing import Any, Dict, Optional
from uuid import UUID

from jinja2 import Environment, StrictUndefined, Template
from jinja2.exceptions import TemplateSyntaxError

from integrations.os2sync.config import loggername as _loggername
from integrations.os2sync.config import get_os2sync_settings

logger = logging.getLogger(_loggername)


class FieldTemplateSyntaxError(Exception):
    pass


class FieldTemplateRenderError(Exception):
    pass


class FieldRenderer:
    """Render Jinja templates defined in app settings"""

    def __init__(self, config: Dict[str, str]):
        """Prepare the field renderer, parsing all Jinja templates defined in
        `config`.

        :param config: dictionary, usually `settings.os2sync_templates`
        """

        # Configure Jinja environment to raise exception on unknown variables
        self._env = Environment(undefined=StrictUndefined)

        def _load_template(key, source):
            try:
                template = self._env.from_string(source)
            except TemplateSyntaxError as e:
                raise FieldTemplateSyntaxError(
                    "syntax error in template %r (source=%r)" % (key, source)
                ) from e
            return template

        # Instantiate all Jinja templates found in config, and map them to
        # their config key
        self._template_fields: Dict[str, Template] = {
            key: _load_template(key, source) for key, source in config.items()
        }

    def render(self, key: str, context: Dict[str, Any], fallback: Any = None) -> str:
        """Render a field template given by `key` using `context`.

        :param key: config key specifying the template to render
        :param context: dictionary used to render template
        :param fallback: value used as fallback in case no template matches `key`
        :raises FieldTemplateRenderError: if template rendering fails
        """

        template = self._template_fields.get(key)

        if not template:
            return fallback

        try:
            return template.render(**context)
        except Exception as e:
            raise FieldTemplateRenderError(
                "could not render template %r (context=%r)" %
                (key, context)
            ) from e


class Entity:
    """Base class for modelling entities defined in the OS2Sync REST API"""

    def __init__(
        self,
        context: Dict[str, Any],
        settings: Dict[str, str],
    ):
        """Configure field renderer for this entity.

        :param context: template rendering context, usually a MO JSON response
        :param settings: app settings, defaults to `os2sync.config.settings`
        """

        self.context = context
        self.settings = settings
        self.field_renderer = FieldRenderer(
            self.settings.os2sync_templates
        )

    def to_json(self) -> Dict[str, Any]:
        """Return a dictionary suitable for inclusion in a JSON payload."""

        raise NotImplementedError("must be implemented by subclass")


class User(Entity):
    """Models a `User` entity in the OS2Sync REST API"""

    def __init__(
        self,
        context: Dict[str, Any],
        settings: Optional[Dict[str, str]] = None,
    ):
        super().__init__(context, settings=settings)
        assert isinstance(context["uuid"], (UUID, str))
        assert isinstance(context["person"], Person)
        assert isinstance(context["candidate_user_id"], (str, type(None)))
        self.context.setdefault("user_key", context["person"].context["user_key"])

    def to_json(self) -> Dict[str, Any]:
        if self.context["candidate_user_id"] is not None:
            # If an AD BVN is available, always use that
            user_id = self.context["candidate_user_id"]
        else:
            # Otherwise, use the "person.user_id" template if available,
            # falling back to the MO user UUID if not.
            user_id = self.field_renderer.render(
                "person.user_id", self.context, fallback=self.context["uuid"],
            )

        person = self.context["person"]

        return {
            "Uuid": self.context["uuid"],
            "UserId": user_id,
            "Person": person.to_json(),
            "Positions": [],
        }


class Person(Entity):
    """Models a `Person` entity in the OS2Sync REST API"""

    def to_json(self) -> Dict[str, Any]:
        if self.settings.os2sync_xfer_cpr:
            cpr = self.context.get("cpr_no")
            if not cpr:
                logger.warning("no 'cpr_no' for user %r", self.context["uuid"])
            else:
                logger.debug("transferring CPR for user %r", self.context["uuid"])
        else:
            cpr = None
            logger.debug("not configured to transfer CPR")

        return {
            "Name": self.field_renderer.render(
                "person.name", self.context, fallback=self.context["name"],
            ),
            "Cpr": cpr,
        }



import unittest

from parameterized import parameterized

from ..utils import AttrDict


class TestDotMapping(unittest.TestCase):
    @parameterized.expand(
        [
            (0,),
            ("value",),
            (3.14,),
        ]
    )
    def test_getattr_static(self, value):
        attr_dict = AttrDict({"status": value})
        self.assertEqual(attr_dict.status, value)

    @parameterized.expand(
        [
            (0,),
            ("value",),
            (3.14,),
        ]
    )
    def test_setattr_static(self, value):
        attr_dict = AttrDict({})
        attr_dict.status = value
        self.assertEqual(attr_dict.status, value)

    @parameterized.expand(
        [
            (0,),
            ("value",),
            (3.14,),
        ]
    )
    def test_delattr_static(self, value):
        attr_dict = AttrDict({"status": value})
        del attr_dict.status
        self.assertEqual(attr_dict, {})

    @parameterized.expand(
        [
            (
                "exit_code",
                0,
            ),
            (
                "key",
                "value",
            ),
            (
                "PI",
                3.14,
            ),
        ]
    )
    def test_getattr_dynamic(self, key, value):
        attr_dict = AttrDict({key: value})
        self.assertEqual(getattr(attr_dict, key), value)

    @parameterized.expand(
        [
            (
                "exit_code",
                0,
            ),
            (
                "key",
                "value",
            ),
            (
                "PI",
                3.14,
            ),
        ]
    )
    def test_setattr_dynamic(self, key, value):
        attr_dict = AttrDict({})
        setattr(attr_dict, key, value)
        self.assertEqual(attr_dict[key], value)

    @parameterized.expand(
        [
            (
                "exit_code",
                0,
            ),
            (
                "key",
                "value",
            ),
            (
                "PI",
                3.14,
            ),
        ]
    )
    def test_delattr_dynamic(self, key, value):
        attr_dict = AttrDict({key: value})
        delattr(attr_dict, key)
        self.assertEqual(attr_dict, {})



import datetime
import logging
from operator import itemgetter
from typing import Dict
from typing import List
from typing import Optional
from uuid import UUID

import click
from click_option_group import optgroup
from click_option_group import RequiredMutuallyExclusiveOptionGroup
from more_itertools import only
from mox_helpers.mox_util import ensure_class_in_lora
from os2mo_helpers.mora_helpers import MoraHelper
from ra_utils.load_settings import load_settings
from tqdm import tqdm

import constants
from . import payloads
from .ad_reader import ADParameterReader

# Set up a real logger!
logger = logging.getLogger("ImportADGroup")

LOG_LEVEL = logging.DEBUG
LOG_FILE = "external_ad_users.log"

for name in logging.root.manager.loggerDict:
    if name in ("ImportADGroup", "AdReader", "mora-helper", "AdCommon"):
        logging.getLogger(name).setLevel(LOG_LEVEL)
    else:
        logging.getLogger(name).setLevel(logging.ERROR)

logging.basicConfig(
    format="%(levelname)s %(asctime)s %(name)s %(message)s",
    level=LOG_LEVEL,
    filename=LOG_FILE,
)


class ADMOImporter(object):
    def __init__(self):

        self.settings = load_settings()
        self.root_ou_uuid = self.settings["integrations.ad.import_ou.mo_unit_uuid"]
        self.helper = MoraHelper(hostname=self.settings["mora.base"], use_cache=False)
        self.org_uuid = self.helper.read_organisation()

        self.ad_reader = ADParameterReader()
        self.ad_reader.cache_all(print_progress=True)

        its = self.helper.read_it_systems()
        AD_its = only(filter(lambda x: x["name"] == constants.AD_it_system, its))
        self.AD_it_system_uuid = AD_its["uuid"]

    def _find_or_create_unit_and_classes(self):
        """
        Find uuids of the needed unit and classes for the import. If any unit or
        class is missing from MO, it will be created.The function returns a dict
        containg uuids needed to create users and engagements.
        """
        # TODO: Add a dynamic creation of classes
        job_type, _ = ensure_class_in_lora("engagement_job_function", "Ekstern")
        eng_type, _ = ensure_class_in_lora("engagement_type", "Ekstern")
        org_unit_type, _ = ensure_class_in_lora("org_unit_type", "Ekstern")

        unit = self.helper.read_ou(uuid=self.root_ou_uuid)
        if "status" in unit:  # Unit does not exist
            payload = payloads.create_unit(
                self.root_ou_uuid, "Eksterne Medarbejdere", org_unit_type, self.org_uuid
            )
            logger.debug("Create department payload: {}".format(payload))
            response = self.helper._mo_post("ou/create", payload)
            assert response.status_code == 201
            logger.info("Created unit for external employees")
            logger.debug("Response: {}".format(response.text))

        uuids = {
            "job_function": job_type,
            "engagement_type": eng_type,
            "unit_uuid": self.root_ou_uuid,
        }
        return uuids

    def _find_ou_users_in_ad(self) -> Dict[UUID, List]:
        """
        find users from AD that match a search string in DistinguishedName.
        """
        search_string = self.settings["integrations.ad.import_ou.search_string"]

        def filter_users(user: Dict) -> bool:
            name = user.get("DistinguishedName")
            if name:
                if search_string in name:
                    return True
            return False

        users = list(filter(filter_users, self.ad_reader.results.values()))
        uuids = map(itemgetter("ObjectGUID"), users)
        users_dict = dict(zip(uuids, users))
        return users_dict

    def _create_user(
        self, ad_user: Dict, cpr_field: str, uuid: Optional[str] = None
    ) -> Optional[UUID]:
        """
        Create or update a user in MO using an AD user as a template.
        The user will share uuid between MO and AD.
        :param ad_user: The ad_object to use as template for MO.
        :return: uuid of the the user.
        """
        cpr_raw = ad_user.get(cpr_field)
        if cpr_raw is None:
            return None
        cpr = cpr_raw.replace("-", "")

        payload = payloads.create_user(cpr, ad_user, self.org_uuid, uuid=uuid)
        logger.info("Create user payload: {}".format(payload))
        r = self.helper._mo_post("e/create", payload)
        assert r.status_code == 201
        user_uuid = UUID(r.json())
        logger.info("Created employee {}".format(user_uuid))
        return user_uuid

    def _connect_user_to_ad(self, ad_user: Dict) -> None:
        """Write user AD username to the AD it system"""

        logger.info("Connect user to AD: {}".format(ad_user["SamAccountName"]))

        payload = payloads.connect_it_system_to_user(ad_user, self.AD_it_system_uuid)
        logger.debug("AD account payload: {}".format(payload))
        response = self.helper._mo_post("details/create", payload)
        assert response.status_code == 201
        logger.debug("Added AD account info to {}".format(ad_user["SamAccountName"]))

    def _create_engagement(
        self, ad_user: Dict, uuids: Dict[str, UUID], mo_uuid: UUID = None
    ) -> None:
        """Create the engagement in MO"""
        # TODO: Check if we have start/end date of engagements in AD
        validity = {"from": datetime.datetime.now().strftime("%Y-%m-%d"), "to": None}

        person_uuid = ad_user["ObjectGUID"]
        if mo_uuid:
            person_uuid = mo_uuid

        # TODO: Check if we can use job title from AD
        payload = payloads.create_engagement(
            ad_user=ad_user, validity=validity, person_uuid=person_uuid, **uuids
        )
        logger.info("Create engagement payload: {}".format(payload))
        response = self.helper._mo_post("details/create", payload)
        assert response.status_code == 201
        logger.info("Added engagement to {}".format(ad_user["SamAccountName"]))

    def cleanup_removed_users_from_mo(self) -> None:
        """Remove users in MO if they are no longer found as external users in AD."""
        yesterday = datetime.datetime.now() - datetime.timedelta(days=1)
        users = self._find_ou_users_in_ad()

        mo_users = self.helper.read_organisation_people(self.root_ou_uuid)

        for key, user in mo_users.items():
            if key not in users:
                # This users is in MO but not in AD:
                payload = payloads.terminate_engagement(
                    user["Engagement UUID"], yesterday
                )
                logger.debug("Terminate payload: {}".format(payload))
                response = self.helper._mo_post("details/terminate", payload)
                logger.debug("Terminate response: {}".format(response.text))
                response.raise_for_status()

    def create_or_update_users_in_mo(self) -> None:
        """
        Create users in MO that exist in AD but not in MO.
        Update name of users that has changed name in AD.
        """

        uuids = self._find_or_create_unit_and_classes()
        users = self._find_ou_users_in_ad()
        for AD in self.settings["integrations.ad"]:
            cpr_field = AD["cpr_field"]

            for user_uuid, ad_user in tqdm(
                users.items(), unit="Users", desc="Updating units"
            ):
                logger.info("Updating {}".format(ad_user["SamAccountName"]))
                cpr = ad_user[cpr_field]
                # Sometimes there is a temporary change of cpr in wich the
                # last character is replaced with an 'x'.
                # This user is ignored by the importer
                # until the cpr has been changed back.
                if cpr[-1].lower() == "x":
                    logger.info("Skipped due to 'x' in cpr.")
                    continue

                mo_user = self.helper.read_user(user_cpr=cpr)
                logger.info("Existing MO info: {}".format(mo_user))

                if mo_user:
                    mo_uuid = mo_user.get("uuid")
                else:
                    mo_uuid = self._create_user(ad_user, cpr_field)

                AD_username = self.helper.get_e_itsystems(
                    mo_uuid, self.AD_it_system_uuid
                )

                if not AD_username:
                    self._connect_user_to_ad(ad_user)

                current_engagements = self.helper.read_user_engagement(user=mo_uuid)
                this_engagement = list(
                    filter(
                        lambda x: x.get("org_unit").get("uuid") == uuids["unit_uuid"],
                        current_engagements,
                    )
                )
                if not this_engagement:
                    self._create_engagement(ad_user, uuids, mo_uuid)


@click.command(help="AD->MO user import")
@optgroup.group("Action", cls=RequiredMutuallyExclusiveOptionGroup)
@optgroup.option("--create-or-update", is_flag=True)
@optgroup.option("--cleanup-removed-users", is_flag=True)
@optgroup.option("--full-sync", is_flag=True)
def import_ad_group(**args):
    """
    Command line interface for the AD to MO user import.
    """
    ad_import = ADMOImporter()

    if args.get("create_or_update"):
        ad_import.create_or_update_users_in_mo()

    if args.get("cleanup_removed_users"):
        ad_import.cleanup_removed_users_from_mo()

    if args.get("full_sync"):
        ad_import.create_or_update_users_in_mo()
        ad_import.cleanup_removed_users_from_mo()


if __name__ == "__main__":
    import_ad_group()



import logging
import unittest

from helpers import dummy_settings
from parameterized import parameterized

from integrations.os2sync.config import loggername as _loggername
from integrations.os2sync.templates import FieldTemplateRenderError
from integrations.os2sync.templates import FieldTemplateSyntaxError
from integrations.os2sync.templates import Person
from integrations.os2sync.tests.helpers import MoEmployeeMixin
from integrations.os2sync.tests.helpers import NICKNAME_TEMPLATE


class TestPerson(unittest.TestCase, MoEmployeeMixin):
    @parameterized.expand(
        [
            # mock CPR, os2sync_xfer_cpr, key of expected CPR value, expected log level
            ("0101013333", True, "cpr_no", logging.DEBUG),
            (None, True, "cpr_no", logging.WARNING),
            ("0101013333", False, None, logging.DEBUG),
            (None, False, None, logging.DEBUG),
        ]
    )
    def test_transfer_cpr(self, cpr, flag, expected_key, expected_log_level):
        mo_employee = self.mock_employee(cpr=cpr)
        settings = dummy_settings
        settings.os2sync_xfer_cpr = flag
        person = Person(mo_employee, settings=settings)
        expected_cpr = mo_employee.get(expected_key)
        with self.assertLogs(_loggername, expected_log_level):
            self.assertDictEqual(
                person.to_json(),
                {"Name": mo_employee["name"], "Cpr": expected_cpr},
            )


class TestPersonNameTemplate(unittest.TestCase, MoEmployeeMixin):
    @parameterized.expand(
        [
            (
                {"nickname": False},  # mo employee response kwargs
                "name",  # key of expected value for `Name`
            ),
            (
                {"nickname": True},  # mo employee response kwargs
                "nickname",  # key of expected value for `Name`
            ),
        ]
    )
    def test_template(self, response_kwargs, expected_key):
        mo_employee = self.mock_employee(**response_kwargs)
        person = Person(mo_employee, settings=self._gen_settings(NICKNAME_TEMPLATE))
        self.assertEqual(person.to_json()["Name"], mo_employee[expected_key])

    def test_template_syntax_error_raises_exception(self):
        mo_employee = self.mock_employee()
        with self.assertRaises(FieldTemplateSyntaxError):
            Person(mo_employee, settings=self._gen_settings("{% invalid jinja %}"))

    @parameterized.expand(
        [
            "{{ name|dictsort }}",
            "{{ unknown_variable }}",
        ]
    )
    def test_template_render_failure_raises_exception(self, template):
        mo_employee = self.mock_employee()
        person = Person(mo_employee, settings=self._gen_settings(template))
        with self.assertRaises(FieldTemplateRenderError):
            person.to_json()

    def _gen_settings(self, template):
        settings = dummy_settings
        settings.os2sync_xfer_cpr = True
        settings.os2sync_templates = {"person.name": template}
        return settings



import datetime
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, List, Optional

import click
import fs
from google.cloud import storage
from more_itertools import one
from ra_utils.load_settings import load_settings
from retrying import retry
from smb.base import NotConnectedError


class OpusReaderInterface(ABC):
    @abstractmethod
    def list_opus_files(self) -> List[str]:
        raise NotImplementedError()

    @abstractmethod
    def read_file(self, blob_name: str) -> str:
        raise NotImplementedError()

    def map_dates(self, dump_list) -> Dict[datetime.datetime, str]:
        """Transforms the list of opus_dumps to a dictionary with dates as keys"""
        dumps = {}
        for opus_dump in dump_list:
            date_part = opus_dump.name[4:18]
            export_time = datetime.datetime.strptime(date_part, "%Y%m%d%H%M%S")
            dumps[export_time] = opus_dump
        return dumps

    def read_latest(self):
        all_files = self.list_opus_files()
        latest_date = max(all_files.keys())
        return self.read_file(all_files[latest_date])


class GcloudOpusReader(OpusReaderInterface):
    def __init__(self, settings):
        settings = settings
        bucket_name = settings["integrations.opus.gcloud_bucket_name"]
        self.client = storage.Client()
        self.bucket = storage.Bucket(self.client, bucket_name)

    def list_opus_files(self) -> Dict[datetime.datetime, str]:
        all_files = list(self.client.list_blobs(self.bucket))
        for f in all_files:
            f.name = f.name.replace("production/", "")
        return self.map_dates(all_files)

    def read_file(self, blob) -> str:
        return blob.download_as_text()


retry_args = {
    "stop_max_attempt_number": 7,
    "wait_fixed": 2000,
}


class SMBOpusReader(OpusReaderInterface):
    @retry(**retry_args)
    def __init__(self, settings):
        self.settings = settings
        user = self.settings["integrations.opus.smb_user"]
        password = self.settings["integrations.opus.smb_password"]
        smb_host = self.settings["integrations.opus.smb_host"]
        self.smb_fs = fs.open_fs(f"smb://{user}:{password}@{smb_host}")

    @retry(**retry_args)
    def list_opus_files(self) -> Dict[datetime.datetime, str]:
        all_files = self.smb_fs.glob("*.xml")
        return self.map_dates(x.info for x in all_files)

    @retry(**retry_args)
    def read_file(self, glob) -> str:
        f = one(self.smb_fs.glob(glob.name))
        return self.smb_fs.readtext(f.path)


class LocalOpusReader(OpusReaderInterface):
    def __init__(self, settings):
        self.settings = settings

    def list_opus_files(self) -> Dict[datetime.datetime, Path]:
        dump_path = Path(self.settings["integrations.opus.import.xml_path"])
        return self.map_dates(dump_path.glob("*.xml"))

    def read_file(self, filename) -> str:
        return filename.read_text()


def get_opus_filereader(settings: Optional[Dict] = None) -> OpusReaderInterface:
    """Get the correct opus reader interface based on values from settings."""
    settings = settings or load_settings()
    if settings.get("integrations.opus.gcloud_bucket_name"):
        return GcloudOpusReader(settings)
    if settings.get("integrations.opus.smb_host"):
        return SMBOpusReader(settings)
    return LocalOpusReader(settings)


@click.group()
def cli():
    """CLI for reading opus-files"""
    pass


@cli.command()
def read_last():
    """Read latest opus-file"""
    ofr = get_opus_filereader()
    click.echo(ofr.read_latest())


@cli.command()
def list_files():
    """Show dates of all opus-files"""
    ofr = get_opus_filereader()
    dumps = ofr.list_opus_files()
    dates = sorted(dumps.keys())
    for date in dates:
        click.echo(date)


@cli.command()
@click.option("--date", prompt=None, type=click.DateTime())
def read_file(date):
    """Read opus-file from specific date. If no date is supplied show all available dates"""
    ofr = get_opus_filereader()
    dumps = ofr.list_opus_files()

    if not date:
        dates = sorted(dumps.keys())
        for date in dates:
            click.echo(date)
        click.echo(f'Choose from above and provide as parameter, eg. --date="{date}"')
    else:
        click.echo(ofr.read_file(dumps[date]))


if __name__ == "__main__":
    cli()
